"""
Module for Tracking and Displaying Usage Statistics of LLM Completions and Transcriptions.

This module provides classes and methods to track usage statistics for language model completions
and transcriptions, compute associated costs, and display the information using Streamlit expanders.
It integrates with DynamoDB for persistent storage of daily usage data.

The cost calucation done here is an estimate at best, and misleading at worst. It doesn't currently
have any way to represent a change in model cost, as the only thing tracked is the actual usage itself.
If it is necessary to be able to compute the changing cost of a model over time, that information will
need to be tracked separately.

Classes:
    UsageStats:
        - Tracks input/output tokens, completions, and transcription data by model.
        - Computes costs per model based on usage and model pricing.
        - Provides methods to render usage statistics as Streamlit expanders.

    GlobalUsageTracker (UsageStats, SingletonResource):
        - A singleton resource that globally tracks usage statistics across sessions.

    DailyUsageTracking (UsageStats, DynamoDbResource):
        - Manages daily usage statistics with DynamoDB integration.
        - Provides methods to retrieve or create daily tracking records.
        - Supports querying usage data over a date range.

Usage:
    - Instantiate `UsageStats` to track and compute costs for LLM usage.
    - Use `GlobalUsageTracker` for application-wide usage tracking.
    - Use `DailyUsageTracking` to store and retrieve daily usage statistics from DynamoDB.

Examples:
    # Compute and display completion costs
    usage_stats = UsageStats(...)
    usage_stats.render_completion_cost_as_expander()

    # Get today's usage tracking record from DynamoDB
    memory = DynamoDbMemory(...)
    daily_usage = DailyUsageTracking.get_for_today(memory)

"""

import datetime
from base64 import b64decode
from io import BytesIO
from typing import ClassVar, Optional, Union

import pandas as pd
from boto3.dynamodb.conditions import Key
from pydantic import BaseModel, Field
from simplesingletable import DynamoDbMemory, DynamoDbResource
from simplesingletable.extras.singleton import SingletonResource
from simplesingletable.models import ResourceConfig

from supersullytools.llm.completions import (
    CompletionModel,
    CompletionResponse,
    ImagePromptMessage,
    PromptAndResponse,
    PromptMessage,
)
from supersullytools.utils.media_manager import MediaManager, MediaType
from supersullytools.utils.misc import now_with_dt


class UsageStats(BaseModel):
    # SECTION 1: llm completion tracking / helpers
    input_tokens_by_model: dict[str, int] = Field(default_factory=dict)
    cached_input_tokens_by_model: dict[str, int] = Field(default_factory=dict)
    reasoning_tokens_by_model: dict[str, int] = Field(default_factory=dict)
    output_tokens_by_model: dict[str, int] = Field(default_factory=dict)
    completions_by_model: dict[str, int] = Field(default_factory=dict)

    def compute_cost_per_model(self, model_classes: Optional[list[CompletionModel]] = None):
        if model_classes is None:
            from supersullytools.llm.completions import ALL_MODELS

            model_classes = ALL_MODELS

        # Build a mapping from model identifiers to model classes
        model_class_by_id = {mc.llm_id: mc for mc in model_classes}
        model_class_by_name = {mc.llm: mc for mc in model_classes}
        model_class_lookup = {**model_class_by_id, **model_class_by_name}

        results = {}
        for model in self.completions_by_model:
            model_class: CompletionModel = model_class_lookup.get(model)
            if not model_class:
                continue  # Skip if model_class is not found

            input_tokens = self.input_tokens_by_model.get(model, 0)
            if model_class.cached_input_price_per_1k:
                cached_input_tokens = self.cached_input_tokens_by_model.get(model, 0)
                regular_cost = ((input_tokens - cached_input_tokens) / 1000) * model_class.input_price_per_1k
                cached_cost = (cached_input_tokens / 1000) * model_class.cached_input_price_per_1k
                input_cost = regular_cost + cached_cost
            else:
                input_cost = input_tokens / 1000 * model_class.input_price_per_1k

            output_tokens = self.output_tokens_by_model.get(model, 0)
            output_cost = output_tokens / 1000 * model_class.output_price_per_1k
            results[model] = input_cost + output_cost
        return results

    def render_completion_cost_as_expander(
        self, header="Total Completions: ", model_classes: Optional[list[CompletionModel]] = None, total_in_header=True
    ):
        import streamlit as st

        stats_df = pd.DataFrame(
            {
                "count": self.completions_by_model,
                "input_tokens": self.input_tokens_by_model,
                "cached_input_tokens": self.cached_input_tokens_by_model,
                "output_tokens": self.output_tokens_by_model,
                "cost": self.compute_cost_per_model(model_classes),
            }
        )
        if not stats_df["count"].count():
            completions_label = "No Completions"
            total_cost = 0.0
        else:
            total_count = int(stats_df["count"].sum())
            total_cost = round(stats_df["cost"].sum(), 4)
            completions_label = f"{header}{total_count}"
            if total_in_header:
                completions_label += f"\n\nApprox Cost: ${total_cost}"

        with st.expander(completions_label):
            if not total_in_header and total_cost > 0:
                st.info(f"Approx Completions Cost: ${total_cost}")
            st.dataframe(stats_df.fillna(0))

    # SECTION 2: transcription tracking / helpers
    transcripts_by_model: dict[str, int] = Field(default_factory=dict)
    seconds_transcribed_by_model: dict[str, int] = Field(default_factory=dict)

    def get_minutes_transcribed_by_model(self) -> dict[str, float]:
        return {k: v / 60 for k, v in self.seconds_transcribed_by_model.items()}

    def compute_transcript_cost_per_model(self, model_cost_by_minute: dict[str, float]):
        minutes_transcribed = self.get_minutes_transcribed_by_model()
        results = {}
        for model in self.transcripts_by_model:
            minutes = minutes_transcribed.get(model, 0.0)
            price_per_minute = model_cost_by_minute.get(model, 0.0)
            results[model] = minutes * price_per_minute
        return results

    def render_transcript_cost_as_expander(self, model_cost_by_minute: dict[str, float]):
        import streamlit as st

        cost_per_model = self.compute_transcript_cost_per_model(model_cost_by_minute)
        stats_df = pd.DataFrame(
            {
                "count": self.transcripts_by_model,
                "minutes": self.get_minutes_transcribed_by_model(),
                "cost": cost_per_model,
            }
        )
        if not stats_df["count"].count():
            completions_label = "No Transcripts"
        else:
            total_count = int(stats_df["count"].sum())
            total_minutes = round(stats_df["minutes"].sum(), 2)
            total_cost = round(stats_df["cost"].sum(), 4)
            completions_label = f"Total Transcripts: {total_count} ({total_minutes} min)\n\nApprox Cost: ${total_cost}"

        with st.expander(completions_label):
            st.dataframe(stats_df.fillna(0))


class GlobalUsageTracker(SingletonResource, UsageStats):
    pass


class DailyUsageTracking(DynamoDbResource, UsageStats):
    day: str

    @classmethod
    def get_for_today(cls, memory: DynamoDbMemory, today: Optional[datetime.date] = None, consistent_read: bool = True):
        today = now_with_dt().date()
        # today = today or datetime.date.today()
        key = today.strftime("%Y%m%d")
        if not (existing := memory.get_existing(key, data_class=cls, consistent_read=consistent_read)):
            return memory.create_new(cls, {"day": today.isoformat()}, override_id=key)
        return existing

    @classmethod
    def get_by_date_range(
        cls, memory: DynamoDbMemory, first_date: datetime.date, last_date: datetime.date
    ) -> list["DailyUsageTracking"]:
        first_day = first_date.isoformat()
        last_day = (last_date + datetime.timedelta(days=1)).isoformat()
        return memory.paginated_dynamodb_query(
            key_condition=(Key("gsitype").eq(cls.__name__) & Key("gsitypesk").between(first_day, last_day)),
            index_name="gsitype",
            resource_class=cls,
        )

    def db_get_gsitypesk(self) -> str:
        # override the base method; use created_at instead of updated_at for the type sort key,
        # to enable easy queries for a range of dates off this index (see the classmethod `get_by_date_range`)
        return self.created_at.isoformat()


class TopicUsageTracking(DynamoDbResource, UsageStats):
    topic: str

    @classmethod
    def get_for_topic(cls, memory: DynamoDbMemory, topic: str, consistent_read: bool = True) -> "TopicUsageTracking":
        if not (existing := memory.get_existing(topic, data_class=cls, consistent_read=consistent_read)):
            return memory.create_new(cls, {"topic": topic}, override_id=topic)
        return existing


class SessionUsageTracking(UsageStats):
    completions: list[PromptAndResponse] = Field(default_factory=list)

    def reset(self):
        self.completions = []
        self.input_tokens_by_model = {}
        self.output_tokens_by_model = {}
        self.completions_by_model = {}


TrackerTypes = Union[UsageStats, GlobalUsageTracker, DailyUsageTracking, TopicUsageTracking]


class StoredPromptAndResponse(DynamoDbResource):
    prompt: list[PromptMessage]
    prompt_image_media_ids: dict[int, list[str]] = Field(default_factory=dict)
    response: CompletionResponse
    resource_config: ClassVar[ResourceConfig] = ResourceConfig(compress_data=True)
    source_tag: Optional[str] = None

    @classmethod
    def create_from_prompt_and_response(
        cls,
        par: PromptAndResponse,
        memory: DynamoDbMemory,
        media_manager: Optional[MediaManager],
        source_tag: Optional[str] = None,
    ) -> "StoredPromptAndResponse":
        """Convert image prompt messages into prompt messages; if an optional media manager is supplied, the images will be preserved and IDs attached."""
        prompt_image_media_ids = {}
        new_prompt: list[PromptMessage] = []
        for idx, message in enumerate(par.prompt):
            if isinstance(message, ImagePromptMessage):
                prompt_image_media_ids[idx] = []
                new_prompt.append(PromptMessage(role=message.role, content=message.content))
                if media_manager:
                    for image, format in zip(message.images, message.image_formats):
                        stored_media_obj = media_manager.upload_new_media(
                            source_file_name=f"image_from_completion:{format}",
                            media_type=MediaType.image,
                            file_obj=BytesIO(b64decode(image.encode())),
                            use_gzip=False,
                        )
                        prompt_image_media_ids[idx].append(stored_media_obj.resource_id)

            else:
                new_prompt.append(message)
        return memory.create_new(
            StoredPromptAndResponse,
            {
                "prompt": new_prompt,
                "response": par.response,
                "prompt_image_media_ids": prompt_image_media_ids,
                "source_tag": source_tag,
            },
        )


class CompletionTracker(object):
    def __init__(
        self,
        memory: Optional[DynamoDbMemory],
        trackers: list[TrackerTypes],
        store_prompt_and_response: bool = False,
        store_source_tag: Optional[str] = None,
        store_prompt_images_media_manager: Optional[MediaManager] = None,
        logger=None,
    ):
        self.memory = memory
        self.trackers = trackers
        self.store_prompt_and_response = store_prompt_and_response
        self.store_prompt_images_media_manager = store_prompt_images_media_manager
        self.store_source_tag = store_source_tag
        self.logger = logger

    def fixup_trackers(self):
        for tracker in self.trackers:
            if not tracker.cached_input_tokens_by_model and isinstance(tracker, DynamoDbResource):
                self.memory.update_existing(tracker, {"cached_input_tokens_by_model": {}})

    def track_completion(
        self,
        model: CompletionModel,
        prompt: list[PromptMessage | ImagePromptMessage],
        completion: CompletionResponse,
        override_trackers: Optional[list[TrackerTypes]] = None,
        store_prompt_and_response: Optional[bool] = None,
    ):
        if store_prompt_and_response is None:
            store_prompt_and_response = self.store_prompt_and_response
        par = PromptAndResponse(prompt=prompt, response=completion)
        llm_to_track = model.llm
        trackers = override_trackers or self.trackers
        for tracker in trackers:
            if isinstance(tracker, SessionUsageTracking):
                if llm_to_track in tracker.input_tokens_by_model:
                    tracker.completions_by_model[llm_to_track] += 1
                    tracker.input_tokens_by_model[llm_to_track] += completion.input_tokens
                    tracker.output_tokens_by_model[llm_to_track] += completion.output_tokens
                else:
                    tracker.completions_by_model[llm_to_track] = 1
                    tracker.input_tokens_by_model[llm_to_track] = completion.input_tokens
                    tracker.output_tokens_by_model[llm_to_track] = completion.output_tokens

                if completion.cached_input_tokens:
                    if llm_to_track in tracker.cached_input_tokens_by_model:
                        tracker.cached_input_tokens_by_model[llm_to_track] += completion.cached_input_tokens
                    else:
                        tracker.cached_input_tokens_by_model[llm_to_track] = completion.cached_input_tokens
                tracker.completions.append(par)

            if isinstance(tracker, DynamoDbResource):
                if self.memory is None:
                    if self.logger:
                        self.logger.warning("Unable to track usage without dynamodb memory")
                    continue
                self.memory.increment_counter(tracker, f"completions_by_model.{llm_to_track}")
                self.memory.increment_counter(tracker, f"input_tokens_by_model.{llm_to_track}", completion.input_tokens)
                self.memory.increment_counter(
                    tracker, f"output_tokens_by_model.{llm_to_track}", completion.output_tokens
                )
                if completion.cached_input_tokens:
                    self.memory.increment_counter(
                        tracker, f"cached_input_tokens_by_model.{llm_to_track}", completion.cached_input_tokens
                    )

        if store_prompt_and_response:
            StoredPromptAndResponse.create_from_prompt_and_response(
                par,
                memory=self.memory,
                media_manager=self.store_prompt_images_media_manager,
                source_tag=self.store_source_tag,
            )



import os
from typing import Optional

from simplesingletable import DynamoDbMemory

from supersullytools.llm.completions import CompletionHandler
from supersullytools.llm.trackers import (
    CompletionTracker,
    DailyUsageTracking,
    GlobalUsageTracker,
    SessionUsageTracking,
    TopicUsageTracking,
)
from supersullytools.utils.media_manager import MediaManager


def get_standard_completion_dynamodb_memory(logger=None) -> DynamoDbMemory:
    if logger is None:
        from logzero import logger
    memory = DynamoDbMemory(logger=logger, table_name=os.environ["COMPLETION_TRACKING_DYNAMODB_TABLE"])
    return memory


def get_standard_completion_media_manager(logger=None) -> MediaManager:
    if logger is None:
        from logzero import logger
    return MediaManager(
        bucket_name=os.environ["COMPLETION_TRACKING_BUCKET_NAME"],
        logger=logger,
        dynamodb_memory=get_standard_completion_dynamodb_memory(logger),
        global_prefix="completion_tracking",
    )


def get_standard_completion_handler(
    extra_trackers: Optional[list] = None,
    topics: Optional[list[str]] = None,
    logger=None,
    include_session_tracker: bool = False,
    store_source_tag: Optional[str] = None,
    **kwargs,
) -> CompletionHandler:
    if logger is None:
        from logzero import logger
    trackers = []
    if os.getenv("COMPLETION_TRACKING_DYNAMODB_TABLE"):
        memory = get_standard_completion_dynamodb_memory(logger=logger)
        trackers.append(GlobalUsageTracker.ensure_exists(memory))
        trackers.append(DailyUsageTracking.get_for_today(memory))
        trackers.extend(TopicUsageTracking.get_for_topic(memory, x) for x in (topics or []))
    else:
        memory = None

    if os.getenv("COMPLETION_TRACKING_BUCKET_NAME"):
        media_manager = get_standard_completion_media_manager(logger)
    else:
        media_manager = None

    trackers.extend(extra_trackers or [])
    if include_session_tracker:
        trackers.append(SessionUsageTracking())

    return CompletionHandler(
        logger=logger,
        completion_tracker=CompletionTracker(
            memory=memory,
            trackers=trackers,
            store_prompt_and_response=bool(memory),
            store_prompt_images_media_manager=media_manager,
            store_source_tag=store_source_tag,
            logger=logger,
        ),
        **kwargs,
    )



import datetime
import json
from contextlib import suppress
from enum import Enum, auto
from logging import Logger
from typing import Any, Callable, Literal, Optional, Type, TypeVar

import jsonref
import pytz
from pydantic import BaseModel, computed_field

from supersullytools.llm.completions import (
    CompletionHandler,
    CompletionModel,
    CompletionResponse,
    ImagePromptMessage,
    PromptMessage,
)

PydanticModel = TypeVar("PydanticModel", bound=BaseModel)


class NoParams(BaseModel):
    pass


class AgentTool(BaseModel):
    name: str
    description: Optional[str] = None
    params_model: Type[PydanticModel]
    mechanism: Callable[[PydanticModel], dict | list | str | PydanticModel]
    safe_tool: bool = False

    def invoke_tool(self, params_dict: dict) -> str:
        params = self.params_model.model_validate(params_dict)
        result = self.mechanism(params)
        match result:
            case BaseModel():
                return result.model_dump_json()
            case str():
                return result
            case _:
                return json.dumps(result, default=str)


class StrEnum(str, Enum):
    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[Any]) -> Any:
        return name


class AgentStates(StrEnum):
    initializing = auto()
    ready_for_message = auto()
    received_message = auto()
    pending_tool_use = auto()
    awaiting_tool_approval = auto()
    using_tools = auto()
    error = auto()


class ToolUseModes(StrEnum):
    manual = auto()  # all tools require manual approval
    automatic_safe_only = auto()  # tools marked "safe" will be automatically called, others require approval
    automatic_unsafe = auto()  # all tools will be automatically called


_CT = TypeVar("_CT", bound=AgentTool)


class ToolAndParams(BaseModel):
    reason: str
    tool: _CT
    params: dict

    @computed_field
    @property
    def tool_name(self) -> str:
        return self.tool.name

    def invoke_tool(self) -> str:
        return self.tool.invoke_tool(self.params)

    def validate_params(self) -> PydanticModel:
        return self.tool.params_model.model_validate(self.params)


class MsgVerificationError(RuntimeError):
    def __init__(self, error_msgs: list[str]):
        self.error_msgs = error_msgs


class ChatAgent(object):
    def __init__(
        self,
        agent_description: str,
        logger: Logger,
        completion_handler: CompletionHandler,
        default_completion_model: CompletionModel | str = "GPT 4 Omni Mini",
        tool_profiles: dict[str, list[_CT]] = None,
        initial_tool_profile: Optional[str] = None,
        tool_use_mode: ToolUseModes = ToolUseModes.automatic_safe_only,
        local_timezone_str="America/Los_Angeles",
        user_preferences: Optional[list[str]] = None,
        require_reason: bool = True,
        initial_llm_context: Optional[
            dict
        ] = None,  # copy initial context from here; modify with add_to_context/remove_from_context
        # names of tools to run during the init process; must work with all default parameters
        init_tools_to_use: Optional[list[str]] = None,
        default_max_response_tokens: int = 1000,
        max_consecutive_tool_calls: int = 4,
    ):
        self.agent_description = agent_description
        self.logger = logger
        self.completion_handler = completion_handler
        self._tool_profiles = (tool_profiles or {}).copy()
        self.tool_use_mode = tool_use_mode
        if initial_tool_profile:
            self.active_tool_profile = initial_tool_profile
        else:
            if self._tool_profiles:
                self.active_tool_profile = next(iter(self._tool_profiles.keys()))
            else:
                self.active_tool_profile = None

        # default completion model
        if isinstance(default_completion_model, CompletionModel):
            self.default_completion_model = default_completion_model
        else:
            self.default_completion_model = self.completion_handler.get_model_by_name_or_id(default_completion_model)

        self.pending_tool_calls = None
        self.approved_tool_calls = None
        self.applied_tool_calls = None
        self.applied_tool_call_results = None
        self.chat_history: list[PromptMessage | ImagePromptMessage] = []
        self.reset_history()
        self.current_state = AgentStates.ready_for_message
        self.local_timezone = pytz.timezone(local_timezone_str)
        self.require_reason = require_reason

        self._status_msg = "Initialization Complete"
        self._user_preferences = [x for x in user_preferences] if user_preferences else []
        self._llm_context = {**initial_llm_context} if initial_llm_context else {}
        self.default_max_response_tokens = default_max_response_tokens
        self.max_consecutive_tool_calls = max_consecutive_tool_calls
        self._current_consecutive_tool_calls = 0

        for tool_name in init_tools_to_use or []:
            self.manually_invoke_tool(tool_name, {}, force_pass=True)
            self.run_agent()
            self._add_chat_msg("PASS", role="assistant")
            self.current_state = AgentStates.ready_for_message

        self._chat_start_idx = len(self.chat_history)

    def add_tool_to_active_profile(self, tool: AgentTool):
        try:
            self.get_current_tool_by_name(tool.name)
        except ValueError:
            self._tool_profiles[self.active_tool_profile].append(tool)

    def replace_user_preferences(self, new_preferences: list[str]):
        self._user_preferences = [x for x in new_preferences] if new_preferences else []

    def add_to_context(self, key, value):
        self._llm_context[key] = value

    def remove_from_context(self, key):
        with suppress(KeyError):
            self._llm_context.pop(key)

    def _set_status_msg(self, msg, callback_fn=None):
        if self._status_msg != msg:
            self._status_msg = msg
            if callback_fn:
                callback_fn(msg)

    def get_current_status_msg(self):
        return self._status_msg

    def get_chat_history(
        self, include_system_messages: bool = False, include_function_calls: bool = False
    ) -> list[PromptMessage | ImagePromptMessage]:
        messages = [x.copy() for x in self.chat_history[self._chat_start_idx :]]

        def ai_system_msg(x: PromptMessage):
            if not x.role == "assistant":
                return False
            return x.content in ["CONTINUE", "PASS"]

        if not include_system_messages:
            if include_function_calls:
                # leave tool call results in place
                messages = [
                    x
                    for x in messages
                    if (not x.role == "system" and not ai_system_msg(x)) or "<tool_result>" in x.content
                ]
            else:
                messages = [x for x in messages if not x.role == "system" and not ai_system_msg(x)]
        if not include_function_calls:
            for idx, msg in enumerate(messages):
                msg.content = msg.content.split("<tool>", maxsplit=1)[0]
                messages[idx] = msg
        return messages

    @property
    def working(self) -> bool:
        match self.current_state:
            case AgentStates.ready_for_message:
                return False
            case AgentStates.awaiting_tool_approval:
                return False
            case AgentStates.error:
                return False
            case _:
                return True

    def reset_history(self):
        self.pending_tool_calls = []
        self.approved_tool_calls = []
        self.applied_tool_calls = []
        self.applied_tool_call_results = []
        self.chat_history: list[PromptMessage | ImagePromptMessage] = []

    def run_agent(
        self,
        max_response_tokens: Optional[int] = None,
        override_model: Optional[CompletionModel | str] = None,
        status_callback_fn: Optional[Callable[[str], None]] = None,
    ):
        # run this in a loop
        max_response_tokens = max_response_tokens or self.default_max_response_tokens

        current_state = self.current_state
        self.logger.debug(f"Running agent {current_state=}")
        match self.current_state:
            case AgentStates.ready_for_message:
                self._set_status_msg("Waiting for user message", status_callback_fn)
            case AgentStates.received_message:
                if self.chat_history[-1].role != "assistant":
                    self._set_status_msg("Agent is generating a message", status_callback_fn)
                    response = self._generate_response(
                        max_response_tokens=max_response_tokens, override_model=override_model
                    )
                    if response:
                        if "<tool>" in response:
                            self._current_consecutive_tool_calls += 1
                            self._set_status_msg("Preparing for tool use", status_callback_fn)
                            self.current_state = AgentStates.pending_tool_use
                        else:
                            self._current_consecutive_tool_calls = 0
                            self.current_state = AgentStates.ready_for_message
                        self._add_chat_msg(msg=response, role="assistant")
                else:
                    self.current_state = AgentStates.error
                    self._set_status_msg(
                        "Invalid state -- received_message but last msg in chat_history is from ai!", status_callback_fn
                    )
            case AgentStates.pending_tool_use:
                self.logger.info("Tool use pending, checking if tool approval is required ")
                pending_tools = self.get_pending_tool_calls()
                if not pending_tools:
                    self.logger.info("No tool use pending, returning ready_for_message state")
                    self.current_state = AgentStates.ready_for_message
                    return self.run_agent()
                match self.tool_use_mode:
                    case ToolUseModes.manual:
                        self.logger.info("Manual mode, approval required for all tools")
                        self.current_state = AgentStates.awaiting_tool_approval
                    case ToolUseModes.automatic_safe_only:
                        self.logger.info('Automatic "safe" tool usage mode, checking for unsafe tools')
                        any_unsafe_tools = False
                        for pending_tool in pending_tools:
                            if not pending_tool.tool.safe_tool:
                                any_unsafe_tools = True
                        if any_unsafe_tools:
                            self.logger.info("Tools include unsafe tool, approval required")
                            self.current_state = AgentStates.awaiting_tool_approval
                        else:
                            self.logger.info("Tools do not include unsafe tool usage, no approval required")
                            self.current_state = AgentStates.using_tools
                    case ToolUseModes.automatic_unsafe:
                        self.logger.info("Operating in automatic_unsafe mode; all tool usage approved")
                        self.current_state = AgentStates.using_tools
                    case _:
                        raise ValueError(self.tool_use_mode)
            case AgentStates.awaiting_tool_approval:
                self._set_status_msg("Waiting for tool approval", status_callback_fn)
            case AgentStates.using_tools:
                self.logger.info("Using tools!")
                self._set_status_msg("Beginning tool use", status_callback_fn)
                tools_and_results = []
                for idx, pending_tool in enumerate(self.get_pending_tool_calls()):
                    self._set_status_msg(f'Using Tool "{pending_tool.tool.name}"', status_callback_fn)
                    result = self._handle_tool_usage(pending_tool)
                    self.applied_tool_calls.append(pending_tool)
                    self.applied_tool_call_results.append(result)
                    tools_and_results.append((pending_tool, result))

                msg = "Tool use complete\n"
                for tool, result in tools_and_results:
                    msg += f"<tool_used>{tool.tool_name}</tool_used>\n<tool_result>\n{result}\n</tool_result>\n"

                if self._current_consecutive_tool_calls < self.max_consecutive_tool_calls:
                    msg += (
                        f"This is consecutive tool call number {self._current_consecutive_tool_calls} "
                        f"of {self.max_consecutive_tool_calls} max. You may now use more tools or respond to the user. "
                        "Do not send the tool_result directly; provide relevant information."
                    )
                else:
                    msg += (
                        f"This is your final ({self.max_consecutive_tool_calls}) tool call; "
                        f"you MUST now send a response to the user with no tool calls."
                    )

                self._add_chat_msg(msg, role="system")

                self.logger.info("Tool use completed, sending results to Agent")
                self.current_state = AgentStates.received_message
                self._set_status_msg("Ready to handle tool results", status_callback_fn)
            case AgentStates.initializing:
                pass
            case _:
                raise ValueError(self.current_state)

    def _handle_tool_usage(self, tool_to_use: ToolAndParams) -> str:
        try:
            return tool_to_use.invoke_tool()
        except Exception as e:
            self.logger.warning("Caught exception using tool, providing error to Agent", exc_info=True)
            return "TOOL FAILED!\n" + str(e)

    def get_current_tool_by_name(self, tool_name: str) -> _CT:
        try:
            return next(x for x in self.get_current_tools() if x.name == tool_name)
        except StopIteration:
            raise ValueError(tool_name)

    def message_from_user(self, msg: str | PromptMessage | ImagePromptMessage):
        if not self.current_state == AgentStates.ready_for_message:
            raise RuntimeError("Cannot accept messages at this time")

        is_prompt_message = False
        if isinstance(msg, (PromptMessage, ImagePromptMessage)):
            is_prompt_message = True
            if not msg.role == "user":
                raise RuntimeError("Role must be user")

        self.logger.info("Agent received message")
        self.current_state = AgentStates.received_message
        if is_prompt_message:
            self._append_chat(msg)
        else:
            self._add_chat_msg(msg, role="user")

    def _append_chat(self, msg: PromptMessage | ImagePromptMessage):
        self.chat_history.append(msg)

    def _add_chat_msg(self, msg: str, role: Literal["system", "user", "assistant"] = "user"):
        self._append_chat(PromptMessage(role=role, content=msg))

    def force_add_chat_msg(self, msg: str, role: Literal["system", "user", "assistant"]):
        self._add_chat_msg(msg, role)

    def get_current_tools(self) -> list[_CT]:
        if not self.tool_use_mode:
            return []
        return self._tool_profiles.get(self.active_tool_profile) or []

    def get_simple_completion(
        self,
        msg: str | PromptMessage | ImagePromptMessage | list[str | PromptMessage | ImagePromptMessage],
        model: Optional[CompletionModel] = None,
    ) -> CompletionResponse:
        if not isinstance(msg, list):
            msg = [msg]
        messages = msg
        final_messages = []
        for this_msg in messages:
            if isinstance(this_msg, str):
                final_messages.append(PromptMessage(content=this_msg, role="user"))
            else:
                final_messages.append(this_msg)

        return self.completion_handler.get_completion(
            model=model or self.default_completion_model, prompt=final_messages
        )

    def manually_invoke_tool(self, tool_name, params: dict, force_pass=False):
        if not self.current_state == AgentStates.ready_for_message:
            raise RuntimeError("Cannot use tools at this time")
        tool = self.get_current_tool_by_name(tool_name)

        msg = f"User triggered tool: {tool.name} / {params}"
        if force_pass:
            msg += "; respond PASS and use data later"

        self._add_chat_msg(
            msg=f"<system>{msg}</system>",
            role="system",
        )
        tool_call_str = json.dumps({"name": tool.name, "reason": f"Manually called: {tool.name}", "parameters": params})
        self._add_chat_msg(f"<tool>{tool_call_str}</tool>", role="assistant")
        self.current_state = AgentStates.using_tools

    def get_pending_tool_calls(self) -> list[ToolAndParams]:
        return_data = []
        for pending_tool_call in self.get_pending_tool_calls_data():
            tool_obj = self.get_current_tool_by_name(pending_tool_call["name"])

            return_data.append(
                ToolAndParams(
                    reason=pending_tool_call.get("reason", "Tool usage"),
                    tool=tool_obj,
                    params=pending_tool_call.get("parameters", {}),
                )
            )

        return return_data

    def get_pending_tool_calls_data(self) -> list[dict]:
        if not self.chat_history:
            return []
        if not (self.chat_history[-1].role == "assistant" and "<tool>" in self.chat_history[-1].content):
            return []
        completion = self.chat_history[-1]

        tool_calls = self.extract_tool_calls_from_msg(completion.content)
        return tool_calls

    def extract_tool_calls_from_msg(self, msg: str) -> list[dict]:
        ai_msg, fn_call_str = msg.split("<tool>", maxsplit=1)
        tool_call_strs = [
            this_call_str.split("</tool>", maxsplit=1)[0] for this_call_str in fn_call_str.split("<tool>")
        ]
        tool_calls = [json.loads(x) for x in tool_call_strs]
        return tool_calls

    def approve_pending_tool_usage(self):
        if not self.current_state == AgentStates.awaiting_tool_approval:
            raise RuntimeError("Not currently awaiting approval for tool usage")
        self.logger.info("Pending tool usage approved, moving into using_tools state")
        self.current_state = AgentStates.using_tools

    def reject_pending_tool_usage(self, why: str):
        if not self.current_state == AgentStates.awaiting_tool_approval:
            raise RuntimeError("Not currently awaiting approval for tool usage")
        self.logger.info("Pending tool usage rejected")
        why = why.strip() or "No reason provided..."
        msg = f"The user rejected your tool calls and supplied the following reason:\n\n{why}"
        self._add_chat_msg(msg, role="system")
        self.current_state = AgentStates.received_message

    def retry_last_response(
        self,
        override_last_prompt_msg: PromptMessage | ImagePromptMessage = None,
        override_model: Callable = None,
    ):
        if not self.current_state == AgentStates.ready_for_message:
            raise RuntimeError("Must be in state ready_for_message to retry a response!")
        final_prompt_msg = override_last_prompt_msg or self.chat_history[-2]
        self.chat_history = self.chat_history[:-2] + [final_prompt_msg]
        self.current_state = AgentStates.received_message
        return self.run_agent(override_model)

    def _generate_response(
        self, max_response_tokens: int, max_attempts=3, override_model: Optional[CompletionModel | str] = None
    ) -> str:
        # allow specific code to override the default completion_fn
        if override_model:
            if isinstance(override_model, CompletionModel):
                this_model = override_model
            else:
                this_model = self.completion_handler.get_model_by_name_or_id(override_model)
        else:
            this_model = self.default_completion_model
        now = datetime.datetime.now(tz=self.local_timezone)
        now_fmt = now.strftime("%A, %b %d %Y %I:%M %p")

        prompt = self.build_tool_usage_prompt(
            prompt=self.agent_description, available_functions=self.get_current_tools()
        )

        prompt += (
            "\n\n---\n\n"
            "To begin, simply use the `BeginChatOperation` tool"
            ' with the startup phrase "OrangeCreamsicle" to enable your tools and '
            "indicate you are ready to process user messages."
        )
        response = """
I am ready for user messages.
<tool>
{
  "name": "BeginChatOperation",
  "reason": "Performing startup task"
  "parameters": {
    "startup_phrase": "OrangeCreamsicle"
  }
}
</tool>
""".strip()

        chat_prefix = [
            PromptMessage(role="user", content=prompt),
            PromptMessage(role="assistant", content=response),
        ]

        final_message: PromptMessage | ImagePromptMessage = self.chat_history[-1].model_copy()

        ephemeral_context = {
            "current_local_time": now_fmt,
            "user_preference_notes": self._user_preferences,
        }
        system_context = {**self._llm_context, **ephemeral_context}
        system_context_str = json.dumps(system_context, default=str, indent=2)
        msg_content = (
            f"<message_from_user>\n{final_message.content}\n</message_from_user>"
            f"\n<system_context>This section provides data injected automatically by the system at runtime."
            f"\n{system_context_str}"
            f"\n</system_context>"
        )
        final_message.content = msg_content

        chat_prompt = chat_prefix + self.chat_history[:-1] + [final_message]

        attempt_num = 1
        bad_responses = []
        while attempt_num <= max_attempts:
            self.logger.info(f"Generating completion, attempt {attempt_num} of {max_attempts}")
            response = self.completion_handler.get_completion(
                model=this_model, prompt=chat_prompt, max_response_tokens=max_response_tokens
            )
            try:
                self._verify(response.content)
            except MsgVerificationError as e:
                self.logger.warning("Generated response failed verification", exc_info=True)
                self.logger.debug(response)
                bad_responses.append(response)
                attempt_num += 1

                if attempt_num <= max_attempts:
                    error_fmt = ", ".join(e.error_msgs)
                    error_msg = f"<system>Error with tool calls: {error_fmt}</system>"
                    chat_prompt += [
                        PromptMessage(role="assistant", content=response.content),
                        PromptMessage(role="system", content=error_msg),
                    ]
                    self._add_chat_msg(response.content, "assistant")
                    self._add_chat_msg(error_msg, "system")
            else:
                # verified with no error, so break out
                break
        else:  # if we didn't hit a break before running out of attempts
            self.logger.error("Failed to get a valid response within max_attempts value!")
            raise ValueError("Bad AI Response")
        return response.content

    def _verify(self, msg: str):
        error_msgs = []
        if "<tool>" in msg:
            if self._current_consecutive_tool_calls >= self.max_consecutive_tool_calls:
                error_msgs.append(
                    "Limit of consecutive tool calls with no user "
                    "response has been reached; must return a message with no tool calls."
                )
            if self.require_reason:
                required_keys = {"name", "reason"}
            else:
                required_keys = {"name"}
            try:
                tool_calls = self.extract_tool_calls_from_msg(msg)
            except Exception as e:
                error_msgs.append(f"Failed parsing on the provided tool call JSON: {e}")
            else:
                for idx, tool_call in enumerate(tool_calls):
                    for key in required_keys:
                        if key not in tool_call:
                            error_msgs.append(f"tool call {idx+1} missing required field `{key}`")

                        try:
                            self.get_current_tool_by_name(tool_call["name"])
                        except ValueError:
                            all_tool_names = [x.name for x in self.get_current_tools()]
                            error_msgs.append(
                                f"Invalid tool specified {tool_call['name']}; Valid options are: "
                                + ", ".join(all_tool_names)
                            )
        if error_msgs:
            raise MsgVerificationError(error_msgs=error_msgs)
        return True

    @classmethod
    def build_tool_usage_prompt(
        cls, prompt: str, available_functions: list[_CT], preamble_prompt: Optional[str] = None
    ) -> str:
        if available_functions:
            functions_block = json.dumps([cls.tool_description_to_dict(x) for x in available_functions], indent=2)
        else:
            functions_block = "No tools currently available."
        usage_prompt = TOOL_USAGE_PROMPT.replace("TOOLS_BLOCK_HERE", functions_block)
        output_prompt = f"{usage_prompt}\n\n---\n\n{prompt}"
        if preamble_prompt:
            return f"{preamble_prompt}\n\n{output_prompt}"
        return output_prompt

    @staticmethod
    def tool_description_to_dict(tool: _CT) -> dict:
        parameters = jsonref.loads(tool.params_model.schema_json())["properties"]
        with suppress(KeyError):
            for param, details in parameters.items():
                del details["title"]
        output = {"name": tool.name, "parameters": parameters}
        descr = tool.params_model.__doc__ or tool.description or ""
        if descr:
            output["description"] = descr

        return output


TOOL_USAGE_PROMPT = """
In this environment, you have access to tools to achieve your goal.

Use a tool by embedding the appropriate tag and JSON:

<tool>
{
  "name": "$TOOL_NAME",
  "reason": "short explanation of why you are using the tool (e.g., 'Getting upcoming events', 'Hiding date columns')",
  "parameters": {
    "$PARAMETER_NAME": "$PARAMETER_VALUE"
  }
}
</tool>

- **Must** include the wrapper tag around the JSON.
- **May** include a short message before tool usage but **may not** add content after.
- You can call multiple tools in parallel by including multiple `<tool>` blocks.
- You can use additional tools after receiving results (e.g., look up an ID before another operation).
- Use your first response to plan how you will accomplish the task (including any branching logic and multiple steps), then proceed.

Available tools:

<available_tools>
TOOLS_BLOCK_HERE
</available_tools>
""".strip()



from typing import Sequence

import streamlit as st


def check_or_x(value: bool) -> str:
    return "✅" if value else "❌"


def resettable_tabs(name: str, tabs=Sequence[str], session_key_prefix: str = "resettable_tabs_") -> str:
    key = f"{session_key_prefix}{name}"
    for x in range(st.session_state.get(key, 0)):
        st.empty()
    return st.tabs(tabs)


def reset_tab_group(name: str, session_key_prefix: str = "resettable_tabs_"):
    key = f"{session_key_prefix}{name}"
    current = st.session_state.get(key, 0)
    st.session_state[key] = current + 1


def flash_message_after_reload(
    msg: str, toast=False, flash_msgs_session_key="flash_msgs", flash_toasts_session_session_key="flash_toast"
):
    key = flash_toasts_session_session_key if toast else flash_msgs_session_key
    if key not in st.session_state:
        st.session_state[key] = [msg]
    else:
        st.session_state[key].append(msg)


def display_flash_msgs(
    flash_msgs_session_key="flash_msgs", flash_toasts_session_session_key="flash_toast", container=None
):
    if flash_msgs := st.session_state.get(flash_msgs_session_key):
        for msg in flash_msgs:
            if container:
                with container:
                    st.write(msg)
            else:
                st.write(msg)
        st.session_state[flash_msgs_session_key] = []
    if toast_msgs := st.session_state.get(flash_toasts_session_session_key):
        for msg in toast_msgs:
            st.toast(msg)
        st.session_state[flash_toasts_session_session_key] = []


# Simplified CSS for the fixed container
FIXED_CONTAINER_CSS = """
div[data-testid="stVerticalBlockBorderWrapper"]:has(div.fixed-container-simple_fixed_container):not(:has(div.not-fixed-container)){
    background-color: transparent;
    position: fixed;
    width: inherit;
    background-color: inherit;
    top: 100;
    z-index: 999;
}

div[data-testid="stVerticalBlockBorderWrapper"]:has(div.fixed-container-simple_fixed_container):not(:has(div.not-fixed-container)) div[data-testid="stVerticalBlock"]:has(div.fixed-container-simple_fixed_container):not(:has(div.not-fixed-container)) > div[data-testid="element-container"] {
    display: none;
}

div[data-testid="stVerticalBlockBorderWrapper"]:has(div.not-fixed-container):not(:has(div[class^='fixed-container-'])) {
    display: none;
}
""".strip()


def simple_fixed_container():
    key = "simple_fixed_container"
    fixed_container = st.container()
    non_fixed_container = st.container()

    css = FIXED_CONTAINER_CSS
    with fixed_container:
        st.markdown(f"<style>{css}</style>", unsafe_allow_html=True)
        st.markdown(f"<div class='fixed-container-{key}'></div>", unsafe_allow_html=True)
    with non_fixed_container:
        st.markdown("<div class='not-fixed-container'></div>", unsafe_allow_html=True)

    with fixed_container:
        return st.container(border=False)



import re
from datetime import datetime, timezone
from random import choices
from string import ascii_lowercase
from typing import TYPE_CHECKING

from plantuml import PlantUML

if TYPE_CHECKING:
    from pandas import DataFrame
    from pydantic import ValidationError


def camel_to_snake(camel_case):
    snake_case = re.sub(r"(?<!^)(?=[A-Z])", "_", camel_case).lower()
    return snake_case


def now_with_dt():
    return datetime.now(tz=timezone.utc)


def date_id(now=None):
    now = now or now_with_dt()
    return now.strftime("%Y%m%d%H%M%S") + "".join(choices(ascii_lowercase, k=6))


def load_data_from_file(file, replace_nan=True) -> "DataFrame":
    import numpy as np
    import pandas as pd

    if file.type == "text/csv":
        data = pd.read_csv(file, dtype=str, low_memory=False)
    elif file.type == "text/tab-separated-values":
        data = pd.read_csv(file, dtype=str, low_memory=False, sep="\t")
    elif (
        file.type == "application/vnd.ms-excel"
        or file.type == "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    ):
        data = pd.read_excel(file, dtype=str)
    elif file.type == "application/json":
        data = pd.read_json(file, dtype=str)
    elif file.type == "application/octet-stream":  # Assuming this is a Parquet file for now
        data = pd.read_parquet(file, dtype=str)
    else:
        raise RuntimeError(f"Unknown / unsupported file type {file.type}")

    if replace_nan:
        return data.replace({np.nan: None})
    else:
        return data


def format_validation_error(error: "ValidationError") -> str:
    """Parse Pydantic ValidationError and return a formatted markdown error message."""

    messages = ["## Validation Errors\n"]

    for error_item in error.errors():
        loc = " -> ".join(map(str, error_item["loc"]))  # Join the location to create a readable path
        msg = error_item["msg"]
        param = error_item["type"]
        messages.append(f"- **Location:** `{loc}`\n  **Message:** {msg}\n  **Type:** `{param}`\n")

    return "\n".join(messages)


def strip_backticks_and_content_type(input_string):
    # Define a regex pattern to match backticks and content type indicator
    pattern = r"```[a-zA-Z]*\n|```"

    # Remove the matching patterns from the input string
    stripped_string = re.sub(pattern, "", input_string)

    return stripped_string


SVG_R = r"(?:<\?xml\b[^>]*>[^<]*)?(?:<!--.*?-->[^<]*)*(?:<svg|<!DOCTYPE svg)\b"
SVG_RE = re.compile(SVG_R, re.DOTALL)


def is_uml_diagram(input_string):
    # Define regex patterns for UML diagram delimiters and key elements
    uml_delimiters_pattern = r"^@startuml.*@enduml$"
    return re.search(uml_delimiters_pattern, input_string, re.DOTALL)


def render_plantuml(plantuml_text):
    """
    Render PlantUML text into an image.

    Parameters:
    - plantuml_text (str): The PlantUML markup to render.

    Returns:
    - bytes: The image data.
    """
    url = PlantUML(url="http://www.plantuml.com/plantuml/img/")
    return url.get_url(plantuml_text)



"""
media_manager.py

This module written primarily by Chat GPT:

https://chatgpt.com/share/f13a0787-4514-444c-95e3-be2daca4b042

This module provides functionality for managing media files stored in an Amazon S3 bucket,
with metadata stored in Amazon DynamoDB. It includes features for uploading, retrieving,
generating previews, and deleting media files, with optional gzip compression and encryption
support for efficient and secure storage of compressible data like CSV and JSON files.

Classes:
    MediaType: An enumeration of supported media types.
    StoredMedia: A Pydantic model representing the metadata for a stored media file.
    MediaManager: A class responsible for managing media files and their metadata.

Functions:
    resize_image(image: Image, max_size: (int, int) = (200, 200)) -> Image:
        Resizes an image to fit within the specified maximum dimensions while maintaining the aspect ratio.

    generate_image_thumbnail(file_obj: IOBase, size: (int, int) = (200, 200)) -> bytes:
        Generates a thumbnail image from the provided image file object.

    generate_audio_waveform(file_obj: IOBase) -> bytes:
        Generates a waveform image from the provided audio file object.

    generate_pdf_preview(file_obj: IOBase) -> bytes:
        Generates a preview image from the first page of the provided PDF file object.

    generate_video_thumbnail(file_obj: IOBase) -> bytes:
        Generates a thumbnail image from the provided video file object using ffmpeg.

    generate_no_preview_available() -> bytes:
        Generates a "No Preview Available" image for unsupported media types or when preview generation fails.

Dependencies:
    - boto3: For interacting with Amazon S3.
    - gzip: For gzip compression and decompression.
    - matplotlib: For generating audio waveform images.
    - PIL (Pillow): For image processing.
    - pydantic: For data validation and settings management.
    - simplesingletable: For interacting with DynamoDB.
    - smart_open: For reading and writing files from/to S3.
    - pydub: For audio file manipulation (optional).
    - pypdfium2: For PDF file manipulation (optional).
    - cryptography: For encryption and decryption of media files (optional).

Usage:
    # Initialize the MediaManager
    media_manager = MediaManager(bucket_name='your-bucket-name', logger=your_logger, dynamodb_memory=your_dynamodb_memory)

    # Upload a new media file
    with open('path/to/your/file.jpg', 'rb') as file_obj:
        media_manager.upload_new_media('file.jpg', MediaType.image, file_obj)

    # Retrieve a media file's metadata and contents
    metadata, contents = media_manager.retrieve_media_metadata_and_contents('media_id')

    # Delete a media file
    media_manager.delete_media('media_id')

Encryption:
    To use encryption features, the `cryptography` library must be installed. You can generate a Fernet key
    and use it for encryption and decryption as shown below:

    # Generate a Fernet key
    from cryptography.fernet import Fernet
    encryption_key = Fernet.generate_key()

    # Initialize the MediaManager with encryption
    media_manager = MediaManager(bucket_name='your-bucket-name', logger=your_logger, dynamodb_memory=your_dynamodb_memory)

    # Upload a new media file with encryption
    with open('path/to/your/file.jpg', 'rb') as file_obj:
        media_manager.upload_new_media('file.jpg', MediaType.image, file_obj, encryption_key=encryption_key)

    # Retrieve and decrypt a media file's contents
    metadata, contents = media_manager.retrieve_media_metadata_and_contents('media_id', encryption_key=encryption_key)

Conditional Import:
    The cryptography library is imported conditionally. If the library is not available, an ImportError will be
    raised when attempting to use encryption or decryption features.
"""

import gzip
import os
import secrets
import subprocess
import tempfile
from enum import Enum
from io import BytesIO, IOBase
from typing import IO, ClassVar, Optional

import boto3
from humanize import naturalsize
from pydantic import ConfigDict, computed_field
from simplesingletable import DynamoDbMemory, DynamoDbResource

from supersullytools.utils.misc import date_id

try:
    # from cryptography.fernet import Fernet
    from cryptography.hazmat.primitives.ciphers.aead import AESGCM

    CRYPTOGRAPHY_AVAILABLE = True
except ImportError:
    CRYPTOGRAPHY_AVAILABLE = False


class MediaType(str, Enum):
    pdf = "pdf"
    text = "text"
    image = "image"
    audio = "audio"
    archive = "archive"
    video = "video"
    document = "document"  # generic catch-all


class StoredMedia(DynamoDbResource):
    src_filename: Optional[str] = None
    media_type: MediaType
    file_size_bytes: int = None  # raw size of uploaded media file
    storage_size_bytes: int = None  # size of the file being sent to storage (possibly compressed / encrypted)
    preview_size_bytes: int = None  # raw size of the preview file
    preview_storage_size_bytes: int = (
        None  # size of the preview being sent to storage (possibly compressed / encrypted)
    )
    content_gzipped: bool = False
    content_encrypted: bool = False
    preview_encrypted: bool = False

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    @computed_field
    @property
    def file_size(self) -> str:
        if self.file_size_bytes:
            return naturalsize(self.file_size_bytes)
        return ""

    @computed_field
    @property
    def storage_size(self) -> str:
        if self.storage_size_bytes:
            return naturalsize(self.storage_size_bytes)
        return ""

    @computed_field
    @property
    def preview_size(self) -> str:
        if self.preview_size_bytes:
            return naturalsize(self.preview_size_bytes)
        return ""

    @computed_field
    @property
    def preview_storage_size(self) -> str:
        if self.preview_storage_size_bytes:
            return naturalsize(self.preview_storage_size_bytes)
        return ""


class MediaManager:
    """
    MediaManager is responsible for handling the upload, retrieval, preview generation,
    and deletion of media files stored in an S3 bucket, with metadata stored in DynamoDB.
    It supports optional gzip compression for efficient storage of highly compressible data.

    Attributes:
        bucket_name (str): The name of the S3 bucket where media files are stored.
        logger (logging.Logger): Logger instance for logging information and errors.
        dynamodb_memory (DynamoDbMemory): Instance for interacting with DynamoDB to store and retrieve metadata.
        global_prefix (str): A prefix that is pre-pended to the file IDs when reading/writing to S3.

    Methods:
        generate_preview(file_obj: IOBase, media_type: MediaType) -> bytes:
            Generates a preview image for the given media file based on its type.

        list_available_media(num: int = 10, oldest_first: bool = True, pagination_key: Optional[str] = None):
            Lists available media in DynamoDB, optionally paginated and sorted.

        upload_new_media(source_file_name: str, media_type: MediaType, file_obj: IOBase, use_gzip: bool = False) -> StoredMedia:
            Uploads a new media file to S3, generates a preview, and stores metadata in DynamoDB.
            Supports optional gzip compression.

        delete_media(media_id: str) -> None:
            Deletes a media file and its preview from S3 and removes the associated metadata from DynamoDB.
            Handles idempotent deletion by checking for the existence of files before attempting to delete them.

        retrieve_metadata(media_id: str) -> StoredMedia:
            Retrieves the metadata for a given media ID from DynamoDB.

        retrieve_media_metadata_and_contents(media_id: str) -> tuple[StoredMedia, IO[bytes]]:
            Retrieves both the metadata and the content of a media file from S3 and DynamoDB.

        retrieve_media_contents(media_id: str) -> IO[bytes]:
            Retrieves the content of a media file from S3. Automatically decompresses the content if it was stored using gzip.

        retrieve_media_preview(media_id: str) -> bytes:
            Retrieves the preview image of a media file from S3.
    """

    def __init__(self, bucket_name: str, logger, dynamodb_memory, global_prefix: str = ""):
        self.bucket_name = bucket_name
        self.logger = logger
        self.dynamodb_memory: DynamoDbMemory = dynamodb_memory
        self.global_prefix = global_prefix.rstrip("/") + "/" if global_prefix else ""

    def generate_preview(self, file_obj: IOBase, media_type: MediaType) -> bytes:
        try:
            media_type = MediaType[media_type]
        except KeyError:
            raise ValueError(f"Invalid media type: {media_type}. Valid types are: {', '.join(list(MediaType))}")

        try:
            if media_type == MediaType.image:
                return generate_image_thumbnail(file_obj)
            elif media_type == MediaType.audio:
                return generate_audio_waveform(file_obj)
            elif media_type == MediaType.video:
                return generate_video_thumbnail(file_obj)
            elif media_type == MediaType.pdf:
                return generate_pdf_preview(file_obj)
        except:  # noqa
            self.logger.exception("Error generating preview")
        return generate_no_preview_available()

    def list_available_media(self, num: int = 10, oldest_first: bool = True, pagination_key: Optional[str] = None):
        return self.dynamodb_memory.list_type_by_updated_at(
            StoredMedia, ascending=oldest_first, pagination_key=pagination_key, results_limit=num
        )

    @staticmethod
    def _encrypt_contents(file_obj: IOBase, encryption_key: bytes) -> IOBase:
        if not CRYPTOGRAPHY_AVAILABLE:
            raise ImportError("cryptography library is not available. Install it to use encryption features.")

        file_obj.seek(0)
        data = file_obj.read()
        # Encrypt a message
        nonce = secrets.token_bytes(12)  # GCM mode needs 12 fresh bytes every time
        ciphertext = nonce + AESGCM(encryption_key).encrypt(nonce, data, b"")
        encrypted_io = BytesIO(ciphertext)
        encrypted_io.seek(0)
        return encrypted_io

    @staticmethod
    def _decrypt_contents(file_obj: IOBase, encryption_key: bytes) -> IOBase:
        if not CRYPTOGRAPHY_AVAILABLE:
            raise ImportError("cryptography library is not available. Install it to use decryption features.")

        file_obj.seek(0)
        ciphertext = file_obj.read()
        decrypted_data = AESGCM(encryption_key).decrypt(ciphertext[:12], ciphertext[12:], b"")
        decrypted_io = BytesIO(decrypted_data)
        decrypted_io.seek(0)
        return decrypted_io

    def upload_new_media(
        self,
        source_file_name: str,
        media_type: MediaType,
        file_obj: IOBase,
        use_gzip: bool = False,
        encryption_key: Optional[bytes] = None,
        encrypt_preview: bool = True,
    ) -> StoredMedia:
        from smart_open import open as smart_open

        try:
            media_type = MediaType[media_type]
        except KeyError:
            raise ValueError(f"Invalid media type: {media_type}. Valid types are: {', '.join(list(MediaType))}")

        upload_id = date_id()
        prefixed_file_name = "/".join([self.global_prefix, upload_id]).replace("//", "/")
        s3_uri = f"s3://{self.bucket_name}/{prefixed_file_name}"

        try:
            file_obj.seek(0, 2)  # Move to the end of the file to get its size
            raw_file_size_bytes = file_obj.tell()
            file_obj.seek(0)  # Reset the file pointer to the beginning
            if use_gzip:
                compressed_io = BytesIO()
                with gzip.GzipFile(fileobj=compressed_io, mode="wb") as gz:
                    gz.write(file_obj.read())
                compressed_io.seek(0)
                write_obj = compressed_io
            else:
                write_obj = file_obj

            encrypted = False
            if encryption_key:
                encrypted = True
                write_obj = self._encrypt_contents(write_obj, encryption_key)

            write_obj.seek(0, 2)  # Move to the end of the file to get its size
            file_size_bytes = write_obj.tell()
            write_obj.seek(0)  # Reset the file pointer to the beginning

            with smart_open(s3_uri, "wb") as s3_file:
                s3_file.write(write_obj.read())

            self.logger.info(f"Successfully uploaded {source_file_name} to {s3_uri}")
            # Generate and upload preview
            file_obj.seek(0)
            preview_io = BytesIO(self.generate_preview(file_obj, media_type))
            preview_io.seek(0, 2)  # Move to the end of the file to get its size
            raw_preview_size_bytes = preview_io.tell()
            preview_io.seek(0)  # Reset the file pointer to the beginning

            preview_encrypted = False
            if encryption_key and encrypt_preview:
                preview_encrypted = True
                preview_io = self._encrypt_contents(preview_io, encryption_key)

            preview_io.seek(0, 2)  # Move to the end of the file to get its size
            preview_file_size_bytes = preview_io.tell()
            preview_io.seek(0)  # Reset the file pointer to the beginning

            preview_s3_uri = f"{s3_uri}_preview"
            with smart_open(preview_s3_uri, "wb") as s3_preview_file:
                s3_preview_file.write(preview_io.read())

            self.logger.info(f"Successfully uploaded preview for {source_file_name} to {preview_s3_uri}")

            # Create and store the media metadata
            metadata = self.dynamodb_memory.create_new(
                StoredMedia,
                {
                    "src_filename": source_file_name,
                    "media_type": media_type,
                    "file_size_bytes": raw_file_size_bytes,
                    "storage_size_bytes": file_size_bytes,
                    "preview_size_bytes": raw_preview_size_bytes,
                    "preview_storage_size_bytes": preview_file_size_bytes,
                    "content_gzipped": use_gzip,
                    "content_encrypted": encrypted,
                    "preview_encrypted": preview_encrypted,
                },
                override_id=upload_id,
            )
            return metadata
        except Exception as e:
            self.logger.exception(f"Failed to upload {source_file_name} to {s3_uri}: {str(e)}")
            raise

    def delete_media(self, media_id: str) -> None:
        metadata = self.retrieve_metadata(media_id)  # Ensure the media exists
        prefixed_file_name = "/".join([self.global_prefix, media_id]).replace("//", "/")
        s3_uri = f"{self.bucket_name}/{prefixed_file_name}"
        preview_s3_uri = f"{s3_uri}_preview"

        s3_client = boto3.client("s3")

        try:
            # Idempotent deletion of the main file from S3
            try:
                s3_client.delete_object(Bucket=self.bucket_name, Key=prefixed_file_name)
                self.logger.info(f"Successfully deleted {s3_uri}")
            except s3_client.exceptions.ClientError as e:
                if e.response["Error"]["Code"] == "NoSuchKey":
                    self.logger.info(f"{s3_uri} already deleted or does not exist")
                else:
                    raise

            # Idempotent deletion of the preview file from S3
            try:
                s3_client.delete_object(Bucket=self.bucket_name, Key=f"{prefixed_file_name}_preview")
                self.logger.info(f"Successfully deleted {preview_s3_uri}")
            except s3_client.exceptions.ClientError as e:
                if e.response["Error"]["Code"] == "NoSuchKey":
                    self.logger.info(f"{preview_s3_uri} already deleted or does not exist")
                else:
                    raise

            # Delete the metadata from DynamoDB
            self.dynamodb_memory.delete_existing(metadata)
            self.logger.info(f"Successfully deleted metadata for media ID {media_id}")
        except Exception as e:
            self.logger.exception(f"Failed to delete media ID {media_id}: {str(e)}")
            raise

    def retrieve_metadata(self, media_id: str) -> StoredMedia:
        try:
            metadata = self.dynamodb_memory.read_existing(media_id, StoredMedia)
            self.logger.info(f"Successfully retrieved metadata for media ID {media_id}")
            return metadata
        except Exception as e:
            self.logger.exception(f"Failed to retrieve metadata for media ID {media_id}: {str(e)}")
            raise

    def retrieve_media_metadata_and_contents(
        self, media_id: str, encryption_key: Optional[bytes] = None
    ) -> tuple[StoredMedia, IO[bytes]]:
        metadata = self.retrieve_metadata(media_id)
        contents = self.retrieve_media_contents(media_id, encryption_key, metadata=metadata)
        return metadata, contents

    def retrieve_media_contents(
        self, media_id: str, encryption_key: Optional[bytes] = None, metadata: Optional[StoredMedia] = None
    ) -> IO[bytes]:
        from smart_open import open as smart_open

        metadata = metadata or self.retrieve_metadata(media_id)
        prefixed_file_name = "/".join([self.global_prefix, media_id]).replace("//", "/")
        s3_uri = f"s3://{self.bucket_name}/{prefixed_file_name}"

        if metadata.content_encrypted and not encryption_key:
            raise ValueError("Content encrypted; must supply encryption key")

        try:
            with smart_open(s3_uri, "rb") as s3_file:
                contents = s3_file.read()

            self.logger.info(f"Successfully retrieved contents for media ID {media_id}")

            output_data = BytesIO(contents)

            if metadata.content_encrypted:
                output_data = self._decrypt_contents(file_obj=output_data, encryption_key=encryption_key)

            if metadata.content_gzipped:
                decompressed_io = BytesIO()
                with gzip.GzipFile(fileobj=output_data, mode="rb") as gz:
                    decompressed_io.write(gz.read())
                decompressed_io.seek(0)
                output_data = decompressed_io

            return output_data

        except Exception as e:
            self.logger.exception(f"Failed to retrieve contents for media ID {media_id}: {str(e)}")
            raise

    def retrieve_media_preview(self, media_id: str, encryption_key: Optional[bytes] = None):
        # note: for speed, we do not pull the metadata and check if the preview is encrypted and
        # raise an error if the key isn't supplied (which is done when retrieving contents)
        # if the preview is encrypted and no key is supplied, the preview image will just be busted
        from smart_open import open as smart_open

        prefixed_file_name = "/".join([self.global_prefix, media_id]).replace("//", "/")
        s3_uri = f"s3://{self.bucket_name}/{prefixed_file_name}_preview"

        try:
            with smart_open(s3_uri, "rb") as s3_file:
                contents = s3_file.read()
            self.logger.info(f"Successfully retrieved preview contents for media ID {media_id}")
            if encryption_key:
                contents = self._decrypt_contents(file_obj=BytesIO(contents), encryption_key=encryption_key).read()
            return contents
        except Exception as e:
            self.logger.exception(f"Failed to retrieve preview contents for media ID {media_id}: {str(e)}")
            return generate_no_preview_available()
            # raise


def resize_image(image, max_size: (int, int) = (200, 200)):
    from PIL import Image

    image.thumbnail(max_size, Image.Resampling.LANCZOS)
    return image


# Helper functions
def generate_image_thumbnail(file_obj: IOBase, size=(200, 200)) -> bytes:
    from PIL import Image

    image = Image.open(file_obj)
    image.thumbnail(size)
    thumb_io = BytesIO()
    image.save(thumb_io, format=image.format)
    thumb_io.seek(0)
    return thumb_io.read()


def generate_audio_waveform(file_obj: IOBase) -> bytes:
    from PIL import Image

    try:
        import matplotlib.pyplot as plt
        from pydub import AudioSegment
        from pydub.exceptions import CouldntDecodeError
    except ImportError:
        return generate_no_preview_available()
    try:
        # Load the audio file
        file_obj.seek(0)
        try:
            audio = AudioSegment.from_file(file_obj)
        except CouldntDecodeError as e:
            if "pcm_s4le" in str(e):
                audio = AudioSegment.from_file(file_obj, format="wav", codec="pcm_s16le")
            else:
                raise

        # Get the raw audio data as an array of samples
        samples = audio.get_array_of_samples()

        # Plot the waveform
        plt.figure(figsize=(10, 2))
        plt.plot(samples)
        plt.axis("off")

        # Save the plot to a BytesIO object
        waveform_io = BytesIO()
        plt.savefig(waveform_io, format="png")
        plt.close()
        waveform_io.seek(0)

        image = Image.open(waveform_io)
        image = resize_image(image, max_size=(200, 200))

        resized_io = BytesIO()
        image.save(resized_io, format="PNG")
        resized_io.seek(0)
        return resized_io.read()
    except Exception as e:
        raise Exception(f"Failed to generate audio waveform: {str(e)}")


def generate_pdf_preview(file_obj: IOBase) -> bytes:
    from PIL import Image

    try:
        import pypdfium2 as pdfium
    except ImportError:
        return generate_no_preview_available()
    file_obj.seek(0)
    pdf = pdfium.PdfDocument(file_obj)
    page = pdf[0]
    pil_image = page.render(scale=2).to_pil()
    preview_io = BytesIO()
    pil_image.save(preview_io, format="JPEG")
    preview_io.seek(0)
    image = Image.open(preview_io)
    image = resize_image(image, max_size=(300, 300))

    resized_io = BytesIO()
    image.save(resized_io, format="PNG")
    resized_io.seek(0)
    return resized_io.read()


def generate_video_thumbnail(file_obj: IOBase) -> bytes:
    from PIL import Image
    from smart_open import open as smart_open

    file_obj.seek(0)
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    frame_file = None
    try:
        temp_file.write(file_obj.read())
        temp_file.flush()  # Ensure all data is written to disk
        temp_file.close()

        # Verify the file is correctly written
        if not os.path.exists(temp_file.name):
            raise Exception(f"Temporary file {temp_file.name} was not created successfully.")

        # Prepare the output file for the thumbnail
        frame_file = tempfile.NamedTemporaryFile(delete=False, suffix=".jpg")
        frame_file.close()

        # Run ffmpeg to extract a frame
        result = subprocess.run(
            ["ffmpeg", "-y", "-i", temp_file.name, "-ss", "00:00:01.000", "-vframes", "1", frame_file.name],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Check if ffmpeg succeeded
        if result.returncode != 0:
            error_message = result.stderr.decode("utf-8")
            raise Exception(f"Failed to generate video thumbnail: {error_message}")

        # Load the frame and convert to bytes
        with smart_open(frame_file.name, "rb") as img_file:
            image = Image.open(img_file)
            image = resize_image(image, max_size=(200, 200))
            thumbnail_io = BytesIO()
            image.save(thumbnail_io, format="JPEG")
            thumbnail_io.seek(0)
            return thumbnail_io.read()
    finally:
        # Cleanup temporary files
        os.remove(temp_file.name)
        if frame_file:
            os.remove(frame_file.name)


def generate_no_preview_available() -> bytes:
    return generate_text_image("No Preview Available")


def generate_text_image(text: str, width=200, height=200, font_size=10) -> bytes:
    from PIL import Image, ImageDraw, ImageFont

    # Generate a "No Preview Available" image
    image = Image.new("RGB", (width, height), color=(255, 255, 255))
    draw = ImageDraw.Draw(image)
    try:
        # Use a TrueType font if available
        font = ImageFont.truetype("arial.ttf", size=font_size)
    except IOError:
        # Otherwise, use the default bitmap font
        font = ImageFont.load_default(size=font_size)

    text_bbox = draw.textbbox((0, 0), text, font=font)
    text_width = text_bbox[2] - text_bbox[0]
    text_height = text_bbox[3] - text_bbox[1]
    text_x = (width - text_width) / 2
    text_y = (height - text_height) / 2
    draw.text((text_x, text_y), text, fill="black", font=font)

    thumb_io = BytesIO()
    image.save(thumb_io, format="JPEG")
    thumb_io.seek(0)
    return thumb_io.read()



from abc import ABC
from base64 import b64decode
from datetime import datetime, timezone
from logging import Logger
from typing import TYPE_CHECKING, Literal, Optional, TypeVar, Union

import boto3
import openai
from pydantic import AwareDatetime, BaseModel, computed_field

if TYPE_CHECKING:
    from mypy_boto3_bedrock_runtime import BedrockRuntimeClient
    from openai import Client

    from supersullytools.llm.trackers import CompletionTracker, UsageStats


class PromptMessage(BaseModel):
    role: Literal["system", "user", "assistant"]
    content: str


class ImagePromptMessage(BaseModel):
    role: Literal["user"] = "user"
    content: str
    images: list[str]  # images in base64 encoded format
    image_formats: list[Literal["jpeg", "png"]]


class CompletionModel(BaseModel, ABC):
    provider: str
    make: str
    llm: str
    llm_id: str
    input_price_per_1k: float
    output_price_per_1k: float

    cached_input_price_per_1k: Optional[float] = None
    supports_images: bool


class CompletionResponse(BaseModel):
    content: str
    input_tokens: int
    cached_input_tokens: int = 0
    reasoning_tokens: int = 0
    output_tokens: int
    llm_metadata: CompletionModel
    generated_at: AwareDatetime
    response_metadata: Optional[dict] = None
    stop_reason: str
    completion_time_ms: int

    @computed_field
    @property
    def completion_cost(self) -> float:
        if self.cached_input_tokens and self.llm_metadata.cached_input_price_per_1k:
            input_tokens = self.input_tokens
            cached_input_tokens = self.cached_input_tokens
            regular_cost = ((input_tokens - cached_input_tokens) / 1000) * self.llm_metadata.input_price_per_1k
            cached_cost = (cached_input_tokens / 1000) * self.llm_metadata.cached_input_price_per_1k
            input_cost = regular_cost + cached_cost
        else:
            input_cost = self.input_tokens / 1000 * self.llm_metadata.input_price_per_1k
        output_cost = self.output_tokens / 1000 * self.llm_metadata.output_price_per_1k
        return round(input_cost + output_cost, 4)


class PromptAndResponse(BaseModel):
    prompt: list[PromptMessage | ImagePromptMessage]
    response: CompletionResponse


class OpenAiModel(CompletionModel):
    provider: Literal["OpenAI"] = "OpenAI"
    supports_system_msgs: bool = True


class BedrockCompletionResponse(BaseModel):
    """A helper class to use with parsing bedrock responses from AWS"""

    content: str
    input_tokens: int
    output_tokens: int


class BedrockModel(CompletionModel):
    provider: Literal["AWS Bedrock"] = "AWS Bedrock"


CompletionModelType = TypeVar("CompletionModelType", bound=CompletionModel)


class CompletionHandler:
    def __init__(
        self,
        logger: Logger,
        openai_client: Optional["Client"] = None,
        bedrock_runtime_client: Optional["BedrockRuntimeClient"] = None,
        available_models: Optional[list["CompletionModel"]] = None,
        completion_tracker: Optional["CompletionTracker"] = None,
        debug_output_prompt_and_response=False,
        enable_openai=True,
        enable_bedrock=True,
        default_max_response_tokens: int = 1000,
    ):
        self.logger = logger
        self.enable_openai = enable_openai
        self.enable_bedrock = enable_bedrock

        self.completion_tracker = completion_tracker

        self.openai_client = None
        if self.enable_openai:
            self.openai_client = openai_client or openai.Client()

        self.bedrock_runtime_client = None
        if self.enable_bedrock:
            self.bedrock_runtime_client = bedrock_runtime_client or boto3.client("bedrock-runtime")

        self.debug_output_prompt_and_response = debug_output_prompt_and_response
        if available_models:
            self.available_models = available_models
        else:
            if enable_bedrock and enable_openai:
                self.available_models = DEFAULT_USE_MODELS
            elif enable_bedrock:
                self.available_models = [x for x in DEFAULT_USE_MODELS if isinstance(x, BedrockModel)]
            elif enable_openai:
                self.available_models = [x for x in DEFAULT_USE_MODELS if isinstance(x, OpenAiModel)]
            else:
                # what do?
                raise ValueError("No models specified")

        self.default_max_response_tokens = default_max_response_tokens

    def get_model_by_name_or_id(self, model_name_or_id: str) -> "CompletionModelType":
        try:
            return next(x for x in self.available_models if x.llm == model_name_or_id)
        except StopIteration:
            try:
                return next(x for x in self.available_models if x.llm_id == model_name_or_id)
            except StopIteration:
                raise ValueError(f"No model found {model_name_or_id=}") from None

    def get_completion(
        self,
        model: Union[str, "CompletionModel"],
        prompt: str | list[PromptMessage | ImagePromptMessage],
        max_response_tokens: Optional[int] = None,
        extra_trackers: Optional[list["UsageStats"]] = None,
    ) -> "CompletionResponse":
        if max_response_tokens is None:
            max_response_tokens = self.default_max_response_tokens
        if isinstance(model, str):
            model = self.get_model_by_name_or_id(model)

        if isinstance(prompt, str):
            prompt = [PromptMessage(content=prompt, role="user")]

        match model:
            case OpenAiModel():
                response = self._get_openai_completion(model, prompt, max_response_tokens)
            case BedrockModel():
                response = self._get_bedrock_completion(model, prompt, max_response_tokens)
            case _:
                raise ValueError(model)

        try:
            if self.completion_tracker:
                self.completion_tracker.track_completion(model, prompt, response)
        except Exception:
            self.logger.warning("Error tracking completion! Continuing", exc_info=True)

        try:
            if extra_trackers:
                if self.completion_tracker:
                    self.completion_tracker.track_completion(model, prompt, response, override_trackers=extra_trackers)
                else:
                    self.logger.warning("Extra trackers provided but no completion tracker configured!")
        except Exception:
            self.logger.warning("Error with extra trackers! Continuing", exc_info=True)

        self.logger.debug("Completion generated\n" + response.model_dump_json(indent=2, exclude={"content"}))

        return response

    def _get_bedrock_completion(
        self,
        llm: BedrockModel,
        prompt: list[PromptMessage | ImagePromptMessage],
        max_response_tokens: int,
        temperature=0.0,
    ) -> "CompletionResponse":
        prompt = [] or prompt
        if not self.enable_bedrock:
            raise RuntimeError("Bedrock completions disabled!")
        # boto_input = llm.prepare_bedrock_body(prompt, max_response_tokens)

        self.logger.info(f"Generating Bedrock Completion {llm.llm_id}")

        # Invoke Bedrock API

        if prompt[0].role == "system":
            system_msg = prompt[0]
            prompt = prompt[1:]
        else:
            system_msg = None

        started_at = datetime.now(timezone.utc)
        messages = []
        debug_print_messages = []
        for msg in prompt:
            add_role = "user" if msg.role == "system" else msg.role
            debug_print_messages.append({"role": add_role, "content": [{"text": msg.content}]})
            match msg:
                case PromptMessage():
                    messages.append({"role": add_role, "content": [{"text": msg.content}]})
                case ImagePromptMessage():
                    if not llm.supports_images:
                        raise ValueError("Specified model does not have image prompt support")
                    content = [{"text": msg.content}]
                    for image, image_fmt in zip(msg.images, msg.image_formats):
                        content.append({"image": {"format": image_fmt, "source": {"bytes": b64decode(image)}}})
                    messages.append(
                        {
                            "role": add_role,
                            "content": content,
                        }
                    )
                case _:
                    raise ValueError("Bad prompt type")

        if self.debug_output_prompt_and_response:
            self.logger.debug(f"LLM input:\n{messages}")

        if system_msg:
            system_prompt = [{"text": system_msg.content}]
        else:
            system_prompt = []
        bedrock_response = self.bedrock_runtime_client.converse(
            modelId=llm.llm_id,
            messages=messages,
            system=system_prompt,
            inferenceConfig={"temperature": temperature, "maxTokens": max_response_tokens},
            # additionalModelRequestFields=additional_model_fields,
        )

        # return response
        response_body = bedrock_response["output"]["message"]["content"][0]["text"].strip()
        finished_at = datetime.now(timezone.utc)
        self.logger.info("Generation complete")
        if self.debug_output_prompt_and_response:
            self.logger.debug(f"LLM Response:\n{bedrock_response}")

        token_usage = bedrock_response["usage"]
        input_tokens = token_usage["inputTokens"]
        output_tokens = token_usage["outputTokens"]
        # total_tokens = token_usage["totalTokens"]
        # stop_reason = bedrock_response["stopReason"]

        return CompletionResponse(
            content=response_body,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            llm_metadata=CompletionModel.model_validate(llm, from_attributes=True),
            generated_at=finished_at,
            completion_time_ms=int((finished_at - started_at).total_seconds() * 1000),
            stop_reason=bedrock_response["stopReason"],
            response_metadata={"usage": bedrock_response["usage"], "metrics": bedrock_response["metrics"]},
        )

    def _get_openai_completion(
        self, llm: OpenAiModel, prompt: list[PromptMessage | ImagePromptMessage], max_response_tokens: int
    ) -> "CompletionResponse":
        if not self.enable_openai:
            raise RuntimeError("OpenAI completions disabled!")
        chat_history = []
        debug_print_chat_history = []
        for msg in prompt:
            debug_print_chat_history.append({"role": msg.role, "content": msg.content})
            if msg.role == "system":
                role = msg.role if llm.supports_system_msgs else "user"
            else:
                role = msg.role
            match msg:
                case PromptMessage():
                    chat_history.append({"role": role, "content": msg.content})
                case ImagePromptMessage():
                    content = [{"type": "text", "text": msg.content}]
                    debug_print_chat_history.append({"role": role, "content": msg.content, "include_image": True})
                    for image, image_fmt in zip(msg.images, msg.image_formats):
                        content.append(
                            {
                                "type": "image_url",
                                "image_url": {"url": f"data:image/{image_fmt};base64,{image}"},
                            }
                        )
                    chat_history.append(
                        {
                            "role": role,
                            "content": content,
                        }
                    )
                case _:
                    raise ValueError("Base prompt type")
        started_at = datetime.now(timezone.utc)

        self.logger.info("Generating Open AI ChatCompletion")
        if self.debug_output_prompt_and_response:
            self.logger.debug(f"LLM input:\n{chat_history}")
        openai_response = self.openai_client.chat.completions.create(
            model=llm.llm_id, messages=chat_history, max_completion_tokens=max_response_tokens
        )
        if self.debug_output_prompt_and_response:
            self.logger.debug(f"LLM response:\n{openai_response}")
        finished_at = datetime.now(timezone.utc)

        return CompletionResponse(
            content=openai_response.choices[0].message.content,
            input_tokens=openai_response.usage.prompt_tokens,
            reasoning_tokens=openai_response.usage.completion_tokens_details.reasoning_tokens or 0,
            cached_input_tokens=openai_response.usage.prompt_tokens_details.cached_tokens or 0,
            output_tokens=openai_response.usage.completion_tokens,
            llm_metadata=CompletionModel.model_validate(llm, from_attributes=True),
            generated_at=finished_at,
            stop_reason=openai_response.choices[0].finish_reason,
            response_metadata=openai_response.model_dump(mode="json", exclude={"choices"}),
            completion_time_ms=int((finished_at - started_at).total_seconds() * 1000),
        )


class Gpt3p5Turbo(OpenAiModel):
    make: str = "OpenAI"
    llm: str = "GPT 3.5 Turbo"
    llm_id: str = "gpt-3.5-turbo"
    input_price_per_1k: float = 0.000500
    output_price_per_1k: float = 0.001500
    supports_images: bool = False


class Gpt4Turbo(OpenAiModel):
    make: str = "OpenAI"
    llm: str = "GPT 4 Turbo"
    llm_id: str = "gpt-4-turbo"
    input_price_per_1k: float = 0.010000
    output_price_per_1k: float = 0.030000
    supports_images: bool = True


class Gpt4Omni(OpenAiModel):
    make: str = "OpenAI"
    llm: str = "GPT 4 Omni"
    llm_id: str = "gpt-4o"
    input_price_per_1k: float = 0.0025
    cached_input_price_per_1k: float = 0.00125
    output_price_per_1k: float = 0.01
    supports_images: bool = True


class Gpt4OmniMini(OpenAiModel):
    make: str = "OpenAI"
    llm: str = "GPT 4 Omni Mini"
    llm_id: str = "gpt-4o-mini"
    input_price_per_1k: float = 0.000150
    cached_input_price_per_1k: float = 0.000075
    output_price_per_1k: float = 0.0006
    supports_images: bool = True


class OpenAIO1Preview(OpenAiModel):
    make: str = "OpenAI"
    llm: str = "o1-preview"
    llm_id: str = "o1-preview"
    input_price_per_1k: float = 0.015
    cached_input_price_per_1k: float = 0.0075
    output_price_per_1k: float = 0.06
    supports_images: bool = False
    supports_system_msgs: bool = False


class OpenAIO1Mini(OpenAiModel):
    make: str = "OpenAI"
    llm: str = "o1-mini"
    llm_id: str = "o1-mini"
    input_price_per_1k: float = 0.003
    cached_input_price_per_1k: float = 0.0015
    output_price_per_1k: float = 0.012
    supports_images: bool = False
    supports_system_msgs: bool = False


class Llama2Chat13B(BedrockModel):
    make: str = "Meta"
    llm: str = "Llama 2 Chat 13B"
    llm_id: str = "meta.llama2-13b-chat-v1"
    input_price_per_1k: float = 0.000750
    output_price_per_1k: float = 0.001000
    supports_images: bool = False


class Llama3p1Instruct8B(Llama2Chat13B):
    make: str = "Meta"
    llm: str = "Llama 3.1 Instruct 8B"
    llm_id: str = "meta.llama3-1-8b-instruct-v1:0"
    input_price_per_1k: float = 0.00022
    output_price_per_1k: float = 0.00022
    supports_images: bool = False


class Llama3p1Instruct70B(Llama2Chat13B):
    make: str = "Meta"
    llm: str = "Llama 3.1 Instruct 70B"
    llm_id: str = "meta.llama3-1-70b-instruct-v1:0"
    input_price_per_1k: float = 0.00099
    output_price_per_1k: float = 0.00099
    supports_images: bool = False


class Llama3p1Instruct405B(Llama2Chat13B):
    make: str = "Meta"
    llm: str = "Llama 3.1 Instruct 405B"
    llm_id: str = "meta.llama3-1-405b-instruct-v1:0"
    input_price_per_1k: float = 0.00532
    output_price_per_1k: float = 0.00532
    supports_images: bool = False


class Claude3Sonnet(BedrockModel):
    make: str = "Anthropic"
    llm: str = "Claude 3 Sonnet"
    llm_id: str = "anthropic.claude-3-sonnet-20240229-v1:0"
    input_price_per_1k: float = 0.003
    output_price_per_1k: float = 0.015
    supports_images: bool = True


class Claude3p5Sonnet(BedrockModel):
    make: str = "Anthropic"
    llm: str = "Claude 3.5 Sonnet"
    llm_id: str = "anthropic.claude-3-5-sonnet-20240620-v1:0"
    input_price_per_1k: float = 0.003
    output_price_per_1k: float = 0.015
    supports_images: bool = True


class Claude3p5SonnetV2(BedrockModel):
    make: str = "Anthropic"
    llm: str = "Claude 3.5 Sonnet V2"
    llm_id: str = "anthropic.claude-3-5-sonnet-20241022-v2:0"
    input_price_per_1k: float = 0.003
    output_price_per_1k: float = 0.015
    supports_images: bool = True


class Claude3Haiku(Claude3Sonnet):
    make: str = "Anthropic"
    llm: str = "Claude 3 Haiku"
    llm_id: str = "anthropic.claude-3-haiku-20240307-v1:0"
    input_price_per_1k: float = 0.00025
    output_price_per_1k: float = 0.00125
    supports_images: bool = True


class Claude3p5Haiku(Claude3Sonnet):
    make: str = "Anthropic"
    llm: str = "Claude 3.5 Haiku"
    llm_id: str = "anthropic.claude-3-5-haiku-20241022-v1:0"
    input_price_per_1k: float = 0.001
    output_price_per_1k: float = 0.005
    supports_images: bool = True


class Claude3Opus(Claude3Sonnet):
    make: str = "Anthropic"
    llm: str = "Claude 3 Opus"
    llm_id: str = "anthropic.claude-3-opus-20240229-v1:0"
    input_price_per_1k: float = 0.015
    output_price_per_1k: float = 0.075
    supports_images: bool = True


class Mistral7B(BedrockModel):
    make: str = "Mistral AI"
    llm: str = "Mistral 7B Instruct"
    llm_id: str = "mistral.mistral-7b-instruct-v0:2"
    input_price_per_1k: float = 0.000150
    output_price_per_1k: float = 0.000200
    supports_images: bool = False


class Mixtral8x7B(Mistral7B):
    make: str = "Mistral AI"
    llm: str = "Mixtral 8x7B Instruct"
    llm_id: str = "mistral.mixtral-8x7b-instruct-v0:1"
    input_price_per_1k: float = 0.000450
    output_price_per_1k: float = 0.000700
    supports_images: bool = False


ALL_MODELS = [
    Gpt3p5Turbo(),
    Gpt4Omni(),
    Gpt4OmniMini(),
    Gpt4Turbo(),
    OpenAIO1Preview(),
    OpenAIO1Mini(),
    Gpt4Turbo(),
    Llama2Chat13B(),
    Llama3p1Instruct8B(),
    Llama3p1Instruct70B(),
    Llama3p1Instruct405B(),
    Mistral7B(),
    Mixtral8x7B(),
    Claude3Haiku(),
    Claude3p5Haiku(),
    Claude3Sonnet(),
    Claude3p5Sonnet(),
    Claude3p5SonnetV2(),
    Claude3Opus(),
]

DEFAULT_USE_MODELS = [
    Gpt4Omni(),
    Gpt4OmniMini(),
    Gpt4Turbo(),
    OpenAIO1Preview(),
    OpenAIO1Mini(),
    Claude3Haiku(),
    Claude3p5Haiku(),
    Claude3p5Sonnet(),
    Claude3p5SonnetV2(),
    Claude3Opus(),
]



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta
from decimal import Decimal
from logging import Logger
from typing import ClassVar, Generic, MutableMapping, Optional, Type, TypeVar

import streamlit as st
from humanize import precisedelta
from pydantic import BaseModel, ConfigDict, Field
from simplesingletable import DynamoDbMemory, DynamoDbResource, DynamoDbVersionedResource
from simplesingletable.models import ResourceConfig

from ..utils.misc import date_id

T = TypeVar("T", bound=BaseModel)


class StreamlitSessionBase(BaseModel):
    """Represents a user session in a Streamlit application with an expiration mechanism.

    Subclass this and add attributes for whatever you need to track in the user session.

    Attributes:
        session_id (str): Unique identifier for the session, generated using a custom method.
        expires_at (Optional[Decimal]): Timestamp at which the session expires.

    Methods:
        expires_in: Returns a string indicating the time remaining until session expiration.
        save_to_session_state: Saves session data to Streamlit's session state for persistence.
    """

    session_id: str = Field(default_factory=date_id)
    expires_at: Optional[Decimal] = None

    @property
    def expires_in(self) -> str:
        if self.expires_at:
            now = datetime.utcnow()
            expiration = datetime.fromtimestamp(float(self.expires_at))

            time_remaining = expiration - now
            if time_remaining.total_seconds() < 0:
                return f"expired {precisedelta(time_remaining)} ago"
            else:
                return f"expires in {precisedelta(time_remaining)}"

        else:
            return ""

    @property
    def is_expired(self) -> bool:
        if self.expires_at:
            now = datetime.utcnow()
            expiration = datetime.fromtimestamp(float(self.expires_at))
            time_remaining = expiration - now
            return time_remaining.total_seconds() < 0
        else:
            return False

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        self.save_to_session_state()

    def save_to_session_state(self):
        datakey = self.__class__.__name__
        if datakey not in st.session_state:
            st.session_state[datakey] = {}
        for k, v in self.model_dump().items():
            st.session_state[datakey][k] = v

    # model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")


SessionType = TypeVar("SessionType", bound=StreamlitSessionBase)


@dataclass
class SessionManagerInterface(ABC, Generic[SessionType]):
    """Defines an interface for managing user sessions in a Streamlit application.

    Methods to be implemented:
        persist_session: Persists the given session.
        get_session: Retrieves a session based on the session ID.
        get_session_model: Returns the session model class.
        get_query_param_name: Returns the query parameter name for the session.
        set_session_expiration: Sets the expiration for a session.
        init_session: Initializes a session, either by creating a new one or loading an existing one.
        clear_session_data: Clears session data from the session state.
        switch_session: Switches to a different session based on the provided session ID.
    """

    logger: Logger
    model_type: Type[SessionType]
    query_param_name: Optional[str] = None

    @abstractmethod
    def persist_session(self, session: SessionType):
        pass

    @abstractmethod
    def get_session(self, session_id: str) -> Optional[SessionType]:
        pass

    # @abstractmethod
    # def get_session_model(self) -> Type[SessionType]:
    #     pass
    #
    # @abstractmethod
    # def get_query_param_name(self) -> str:
    #     pass

    def get_session_model(self) -> Type[SessionType]:
        return self.model_type

    def get_query_param_name(self) -> str:
        if not self.query_param_name:
            return self.get_session_model().__name__
        return self.query_param_name

    def set_session_expiration(self, session: SessionType, expiration: datetime | timedelta):
        match expiration:
            case datetime():
                expires_at = expiration.timestamp()
            case timedelta():
                expires_at = (datetime.utcnow() + expiration).timestamp()
            case _:
                raise ValueError("Invalid type for expiration")
        expires_at = Decimal(str(expires_at).split(".")[0])
        session.expires_at = expires_at
        self.persist_session(session)
        return session

    def init_session(
        self, expiration: Optional[datetime | timedelta] = None, auto_extend_session_expiration=False
    ) -> SessionType:
        datakey = self.get_session_model().__name__
        query_session = st.query_params.get(self.get_query_param_name())
        if query_session:
            if st.session_state.get(datakey, {}).get("session_id") != query_session:
                # no, the requested session isn't loaded -- clear out any existing session data
                self.logger.info("Clearing outdated session")
                self.clear_session_data()
        if datakey not in st.session_state:
            st.session_state[datakey] = {}

        session: Optional[T] = None
        if st.session_state[datakey].get("session_id"):
            # since we have a session id in the session_state already, we've done an init and can load the data
            session = self.get_session_model().model_validate(st.session_state[datakey])
        elif query_session:
            self.logger.info("Loading session from query param")
            session = self.get_session(query_session)
            if not session:
                self.logger.warning(f"No session matching query param found {query_session=}")
                try:
                    del st.query_params[self.get_query_param_name()]
                except KeyError:
                    pass

        if not session:
            self.logger.info(f"Starting new session {datakey=}")
            if expiration:
                match expiration:
                    case datetime():
                        expires_at = expiration.timestamp()
                    case timedelta():
                        expires_at = (datetime.utcnow() + expiration).timestamp()
                    case _:
                        raise ValueError("Invalid type for expiration")
                expires_at = Decimal(str(expires_at).split(".")[0])
                session = self.get_session_model()(expires_at=expires_at)
            else:
                session = self.get_session_model()()

        session: StreamlitSessionBase

        if session.is_expired:
            self.logger.info("Session is expired; clearing and starting new")
            self.clear_session_data()
            session = self.init_session(expiration=expiration)
        else:
            if expiration and auto_extend_session_expiration:
                match expiration:
                    case datetime():
                        expires_at = expiration.timestamp()
                    case timedelta():
                        expires_at = (datetime.utcnow() + expiration).timestamp()
                    case _:
                        raise ValueError("Invalid type for expiration")
                expires_at = Decimal(str(expires_at).split(".")[0])
                session.expires_at = expires_at
        session.save_to_session_state()
        return session

    def clear_session_data(self):
        datakey = self.get_session_model().__name__
        st.session_state.pop(datakey, None)
        try:
            del st.query_params[self.get_query_param_name()]
        except KeyError:
            pass

    def switch_session(self, switch_to_session_id):
        incoming_session = self.get_session(switch_to_session_id)
        self.clear_session_data()
        incoming_session.save_to_session_state()


class InMemorySessionManager(SessionManagerInterface[SessionType]):
    """Manager with no persistence at all."""

    def __init__(
        self,
        model_type: Type[SessionType],
        logger,
        memory: Optional[MutableMapping] = None,
        query_param_name: Optional[str] = None,
    ):
        if memory is None:
            memory = {}
        self.session_store = memory
        self.query_param_name = query_param_name
        self.logger = logger

        if not issubclass(model_type, StreamlitSessionBase):
            raise TypeError("model_type must be a subclass of StreamlitSessionBase")
        self.model_type = model_type

    def persist_session(self, session: SessionType):
        self.session_store[session.session_id] = session
        st.query_params[self.get_query_param_name()] = session.session_id

    def get_session(self, session_id: str) -> Optional[SessionType]:
        self.logger.info("Getting session from memory")
        if db_session := self.session_store.get(session_id):
            return db_session


class DynamoDbSessionManager(SessionManagerInterface[SessionType]):
    """Manages user sessions using DynamoDB as a storage backend in a Streamlit application.

    Attributes:
        logger (Logger): Logger for logging information.
        enable_versioning (bool): Flag to enable versioning for session data.
        _memory (DynamoDbMemory): DynamoDB memory object for data persistence.
        ttl_attribute_name (Optional[str]): Name of the attribute used for time-to-live in DynamoDB.
        _query_param_name (Optional[str]): Query parameter name used in URLs to identify sessions.
        model_type (Type[SessionType]): Type of the session model.

    Methods:
        get_session_model: Returns the session model class.
        get_query_param_name: Returns the query parameter name for the session.
        persist_session: Saves the session to DynamoDB.
        get_session: Retrieves a session from DynamoDB using the session ID.
        get_db_session: Helper method to retrieve a session wrapper from DynamoDB.
    """

    def __init__(
        self,
        model_type: Type[SessionType],
        memory: DynamoDbMemory,
        logger,
        ttl_attribute_name: Optional[str] = None,
        query_param_name: Optional[str] = None,
        enable_versioning: bool = False,
    ):
        self.query_param_name = query_param_name
        self.logger = logger
        self.enable_versioning = enable_versioning
        self._memory = memory
        self.ttl_attribute_name = ttl_attribute_name

        if not issubclass(model_type, StreamlitSessionBase):
            raise TypeError("model_type must be a subclass of StreamlitSessionBase")
        self.model_type = model_type

        if enable_versioning:

            class DbSession(DynamoDbVersionedResource):
                session: model_type
                model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow")

                def to_dynamodb_item(self, v0_object: bool = False) -> dict:
                    base = super().to_dynamodb_item(v0_object)
                    if self.session.expires_at and v0_object and ttl_attribute_name:
                        base[ttl_attribute_name] = self.session.expires_at
                    return base

        else:

            class DbSession(DynamoDbResource):
                session: model_type
                model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow")
                resource_config = ResourceConfig(compress_data=True)

                def to_dynamodb_item(self) -> dict:
                    base = super().to_dynamodb_item()
                    if self.session.expires_at and ttl_attribute_name:
                        base[ttl_attribute_name] = self.session.expires_at
                    return base

        self._db_model = DbSession

    def persist_session(self, session: SessionType):
        existing = self.get_db_session(session.session_id)
        if not existing:
            self._memory.create_new(self._db_model, data={"session": session}, override_id=session.session_id)
        else:
            if existing.session != session:
                self._memory.update_existing(existing, update_obj={"session": session})
        st.query_params[self.get_query_param_name()] = session.session_id

    def get_session(self, session_id: str) -> Optional[SessionType]:
        self.logger.info("Getting session from database")
        if db_session := self.get_db_session(session_id):
            return db_session.session

    # not part of the interface
    def get_db_session(self, session_id: str):
        return self._memory.get_existing(session_id, data_class=self._db_model)



# streamlit helpers for ChatAgent
import datetime
import json
import time
from base64 import b64decode, b64encode
from typing import TYPE_CHECKING, Callable, Optional

import streamlit as st
from humanize import precisedelta
from pydantic import BaseModel, ValidationError
from streamlit.runtime.uploaded_file_manager import UploadedFile

from supersullytools.llm.agent import AgentTool, ChatAgent
from supersullytools.llm.completions import CompletionHandler, CompletionModel, ImagePromptMessage, PromptAndResponse
from supersullytools.llm.trackers import SessionUsageTracking, StoredPromptAndResponse
from supersullytools.streamlit.paginator import item_paginator
from supersullytools.utils.misc import format_validation_error

if TYPE_CHECKING:
    from supersullytools.utils.media_manager import MediaManager


class SlashCmd(BaseModel):
    name: str
    description: str
    mechanism: Callable
    refresh_after: bool = False


class ChatAgentUtils(object):
    def __init__(
        self,
        chat_agent: ChatAgent,
        use_system_slash_cmds: bool = True,
        extra_slash_cmds: Optional[dict[str, SlashCmd]] = None,
    ):
        self.chat_agent = chat_agent
        self.use_system_slash_cmds = use_system_slash_cmds
        self._system_slash_cmds = {
            "/help": SlashCmd(
                name="Help",
                description="Display available commands",
                mechanism=self._display_slash_help,
                refresh_after=False,
            ),
        }

        if self.use_system_slash_cmds:
            self._system_slash_cmds["/clear"] = SlashCmd(
                name="Clear Chat",
                description="Clear current chat history",
                mechanism=self.chat_agent.reset_history,
                refresh_after=True,
            )
            self._system_slash_cmds["/tool"] = SlashCmd(
                name="Use a Tool",
                description="Opens a dialog to enable manual tool usage",
                mechanism=self._use_tool_manually,
                refresh_after=False,
            )
            if self.has_session_tracker():
                self._system_slash_cmds["/completions"] = SlashCmd(
                    name="View LLM Completions",
                    description="Opens a dialog to browse LLM completions from the session",
                    mechanism=self._browse_session_completions,
                    refresh_after=False,
                )
        self.extra_slash_cmds = extra_slash_cmds or {}

    def get_session_tracker(self) -> Optional[SessionUsageTracking]:
        if not self.chat_agent.completion_handler.completion_tracker:
            return None
        trackers = self.chat_agent.completion_handler.completion_tracker.trackers
        try:
            return next(x for x in trackers if isinstance(x, SessionUsageTracking))
        except StopIteration:
            return None

    def has_session_tracker(self) -> bool:
        return bool(self.get_session_tracker())

    def _display_slash_help(self):
        output = "### Available Commands\n\n"
        all_commands = {**self.extra_slash_cmds, **self._system_slash_cmds}
        for slash_cmd, obj in all_commands.items():
            obj: SlashCmd
            output += f"* **{obj.name}** (`{slash_cmd}`): {obj.description}\n"
        return output

    def get_completion_model(self, model: Optional[str | CompletionModel] = None) -> CompletionModel:
        if isinstance(model, CompletionModel):
            return model
        return self.chat_agent.completion_handler.get_model_by_name_or_id(model)

    @staticmethod
    def select_llm(
        completion_handler: CompletionHandler, label, default: str = "GPT 4 Omni Mini", key=None, **kwargs
    ) -> CompletionModel:
        completion_handler = completion_handler
        default_model = completion_handler.get_model_by_name_or_id(default)
        return st.selectbox(
            label,
            completion_handler.available_models,
            completion_handler.available_models.index(default_model),
            format_func=lambda x: x.llm,
            key=key,
            **kwargs,
        )

    def select_llm_from_agent(
        self, label, default_override: Optional[str] = None, key=None, **kwargs
    ) -> CompletionModel:
        default_model = (
            self.chat_agent.completion_handler.get_model_by_name_or_id(default_override)
            if default_override
            else self.chat_agent.default_completion_model
        )
        return self.select_llm(
            completion_handler=self.chat_agent.completion_handler,
            label=label,
            default=default_model.llm,
            key=key,
            **kwargs,
        )

    def display_chat_msg(self, msg: str):
        if "<tool>" in msg:
            content, _ = msg.split("<tool>", maxsplit=1)
            content = content.strip()
            try:
                tool_calls = self.chat_agent.extract_tool_calls_from_msg(msg)
            except Exception:
                content += "\n\n<msg contains malformed tool call>"
                tool_calls = []
            if content:
                st.write(content)
            for tc in tool_calls:
                if params := tc.get("parameters"):
                    with st.popover(tc["name"]):
                        st.write(params)
                else:
                    st.caption(f"used `{tc['name']}`")
        elif "<tool_result>" in msg:
            _, result_str = msg.split("<tool_result>", maxsplit=1)
            tool_result_str = [
                this_result_str.split("</tool_result>", maxsplit=1)[0]
                for this_result_str in result_str.split("<tool_result>")
            ]
            tool_results = [x.strip() for x in tool_result_str]
            st.json(tool_results, expanded=0)

        else:
            st.write(msg)

    def _try_handle_system_slash_command(self, command_str: str) -> Optional[tuple[SlashCmd, str]]:
        if command_str in self._system_slash_cmds:
            command: SlashCmd = self._system_slash_cmds[command_str]
            output = command.mechanism()
            if output:
                with st.chat_message("system").container(border=True):
                    st.write(output)
            return command, output

    def _try_handle_extra_slash_cmd(self, command_str: str) -> Optional[tuple[SlashCmd, str]]:
        if command_str in self.extra_slash_cmds:
            command: SlashCmd = self.extra_slash_cmds[command_str]
            output = command.mechanism()
            if output:
                with st.chat_message("system").container(border=True):
                    st.write(output)
            return command, output

    def add_user_message(self, msg: str, images: Optional[list[UploadedFile]] = None) -> bool:
        """Returns true if the streamlit app should reload."""
        if msg.startswith("/"):
            if not (self.use_system_slash_cmds or self.extra_slash_cmds):
                raise RuntimeError("No slash commands available!")
            executed_command = None
            if self.use_system_slash_cmds or msg.startswith("/help"):
                result = self._try_handle_system_slash_command(msg)
                if result:
                    executed_command, _ = result
            if not executed_command and self.extra_slash_cmds:
                result = self._try_handle_extra_slash_cmd(msg)
                if result:
                    executed_command, _ = result
            if not executed_command:
                executed_command, _ = self._try_handle_system_slash_command("/help")
            return executed_command.refresh_after
        else:
            if images:
                prompt = ImagePromptMessage(
                    content=msg,
                    images=[b64encode(image.getvalue()).decode() for image in images],
                    image_formats=["png" if image.name.endswith("png") else "jpeg" for image in images],  # noqa
                )
            else:
                prompt = msg
            self.chat_agent.message_from_user(prompt)
            return True

    def display_chat_and_run_agent(self, include_function_calls):
        num_chat_before = len(self.chat_agent.get_chat_history(include_function_calls=include_function_calls))

        for msg in self.chat_agent.get_chat_history(include_function_calls=include_function_calls):
            with st.chat_message(msg.role):
                if isinstance(msg, ImagePromptMessage):
                    main, images = st.columns((90, 10))
                    with main:
                        self.display_chat_msg(msg.content)
                    for x in msg.images:
                        images.image(b64decode(x.encode()))
                else:
                    self.display_chat_msg(msg.content)

        if self.chat_agent.working:
            with st.status("Agent working...", expanded=True) as status:
                # Define the callback function within the scope of `status`
                def status_callback_fn(message):
                    status.update(label=f"Agent working... {message}", state="running")
                    st.write(message)

                # Run the agent loop, passing the callback function
                while self.chat_agent.working:
                    self.chat_agent.run_agent(status_callback_fn=status_callback_fn)
                    time.sleep(0.05)

                # Final status update when the agent completes
                status.update(label="Agent completed work!", state="complete", expanded=False)

        # output any new messages
        for msg in self.chat_agent.get_chat_history(include_function_calls=include_function_calls)[num_chat_before:]:
            with st.chat_message(msg.role):
                self.display_chat_msg(msg.content)

    def _browse_session_completions(self):
        now = datetime.datetime.now(tz=datetime.timezone.utc)

        @st.dialog("Completions Viewer", width="large")
        def _d():
            if not self.has_session_tracker():
                st.warning("No session tracker found!")

            st.caption("Completions displayed newest-first")
            completions = list(reversed(self.get_session_tracker().completions))

            def _display(idx):
                display_completion(completions[idx], now)

            item_paginator(
                "Completion",
                [x.prompt[-1].content[:25] for x in completions],
                item_handler_fn=_display,
                enable_keypress_nav=True,
                display_item_names=True,
            )

        return _d()

    def _use_tool_manually(self):
        @st.dialog("Trigger Tool", width="large")
        def _d():
            tool_to_use: AgentTool = st.selectbox(
                "Tool to use", self.chat_agent.get_current_tools(), format_func=lambda x: x.name
            )
            tool_dict = self.chat_agent.tool_description_to_dict(tool_to_use)
            if descr := tool_dict.get("description"):
                st.write(descr)
            st.write(tool_dict["parameters"])
            default = {x: None for x in tool_dict["parameters"]}
            params_str = st.text_area("tool params", json.dumps(default, indent=2), height=300)

            try:
                params_obj = tool_to_use.params_model.model_validate_json(params_str)
            except ValidationError as e:
                st.error(format_validation_error(e))
                with st.popover("Invoke Tool", disabled=True):
                    pass
            else:
                invoke_params = json.loads(params_str)
                with st.popover("Invoke Tool"):
                    st.subheader(f"Calling tool {tool_to_use.name}")
                    st.write("Tool Call Parmeters:")
                    st.code(params_obj.model_dump_json(indent=2))
                    if st.button(
                        "Invoke Locally",
                        help="Execute the tool and show the results here, but do to store in session/chat",
                    ):
                        st.write(tool_to_use.invoke_tool(invoke_params))
                    if st.button("Have Agent Invoke"):
                        self.chat_agent.manually_invoke_tool(tool_to_use.name, invoke_params)
                        time.sleep(0.01)
                        st.rerun()

        return _d()


@st.cache_data
def get_media_preview(_media_manager: "MediaManager", media_id):
    return _media_manager.retrieve_media_preview(media_id)


@st.cache_data
def get_stored_media(_media_manager: "MediaManager", media_id):
    return _media_manager.retrieve_media_contents(media_id)


def display_completion(
    par: PromptAndResponse | StoredPromptAndResponse, now, media_manager: Optional["MediaManager"] = None
):
    generated_ago = precisedelta(now - par.response.generated_at, minimum_unit="minutes")
    st.caption(par.response.generated_at.isoformat() + f" ({generated_ago} ago)")

    get_full_image_contents = False
    if media_manager and isinstance(par, StoredPromptAndResponse):
        get_full_image_contents = st.toggle("Retrieve full Image contents")

    is_stored_par = isinstance(par, StoredPromptAndResponse)

    if is_stored_par:
        st.metric("Stored Size", par.get_db_item_size())
        par.resource_config["compress_data"] = False
        st.metric("Uncompressed Size", par.get_db_item_size())
        par.resource_config["compress_data"] = True

    if st.toggle("Show raw"):
        if is_stored_par:
            st.code(par.model_dump_json(indent=2, exclude=par.get_db_resource_base_keys()))
        else:
            st.code(par.model_dump_json(indent=2))
    else:
        st.write("**Prompt**")
        with st.container(border=True):
            if isinstance(par.prompt, str):
                st.code(par.prompt)
            else:
                for idx, msg in enumerate(par.prompt):
                    with st.chat_message(msg.role):
                        if len(msg.content) > 100:
                            with st.popover(msg.content[:100] + " ... ", use_container_width=True):
                                st.write(msg.content)
                        else:
                            st.write(msg.content)
                        if isinstance(msg, ImagePromptMessage):
                            cols = iter(st.columns(len(msg.images)))
                            for x in msg.images:
                                next(cols).image(b64decode(x.encode()))
                        if is_stored_par:
                            if idx in par.prompt_image_media_ids:
                                st.caption("This message included image content")
                                stored_media_ids = par.prompt_image_media_ids[idx]
                                if not stored_media_ids:
                                    st.warning("Images not preserved")
                                    continue
                                if media_manager:
                                    cols = iter(st.columns(len(stored_media_ids)))
                                    for x in stored_media_ids:
                                        if get_full_image_contents:
                                            next(cols).image(get_stored_media(media_manager, x))
                                        else:
                                            with next(cols):
                                                st.image(get_media_preview(media_manager, x))
                                                load_full = st.button("➕", key=x)
                                            if load_full:
                                                st.image(get_stored_media(media_manager, x), use_column_width=True)

        st.write("**Response**")
        with st.container(border=True):
            st.code(par.response.content)

        st.write("**Metadata**")
        with st.container(border=True):
            cols = iter(st.columns(2))
            with next(cols):
                st.json(par.response.llm_metadata.model_dump(mode="json"))
            with next(cols):
                st.json(par.response.model_dump(mode="json", exclude={"content", "llm_metadata", "response_metadata"}))
            st.write("**Raw Provider Response Metadata**")
            st.json(par.response.response_metadata or {})

    st.caption(par.response.generated_at.isoformat() + f" ({generated_ago} ago)")



"""
This module provides tools for performing similarity searches on collections of strings or objects implementing the SimilaritySearchable protocol. It uses the Levenshtein distance to calculate the similarity between search terms and elements in a collection. The module supports customized scoring based on insertion, deletion, and substitution costs, and can handle both exact and partial string matches.

Key components:
- `SimilaritySearchable`: A protocol that requires methods for obtaining the full value and search terms of an object.
- `levenshtein_distance`: Function to compute the edit distance between two strings with customizable operation costs.
- `score_string_similarity`: Calculates a similarity score between a given search term and elements in a collection.
- `MatchedString`: A data class for storing match results with a similarity score.
- `get_top_n_results`: Retrieves the top N similar items to a given search term from a collection.

Usage Example:
    # Define a searchable class
    from dataclasses import dataclass
    from typing import Iterable

    @dataclass
    class MySearchable:
        full_text: str
        terms: list

        def get_full_value(self) -> str:
            return self.full_text

        def get_search_terms(self) -> Iterable[str]:
            return self.terms

    # Create a collection of searchable items
    items = [MySearchable("Example text", ["example", "text"]), "simple text"]

    # Search for the term 'examp'
    results = get_top_n_results(items, "examp", max_results=2)

Limitations:
- Performance: The module is designed for collections of a few thousand items. Its performance with larger datasets has not been extensively tested.
- Memory Usage: The Levenshtein distance calculation can be memory-intensive for very long strings or very large collections.
- Scalability: While suitable for moderate-sized collections, the implementation may require optimization for handling large-scale datasets efficiently.
"""

from dataclasses import dataclass
from functools import partial
from typing import Iterable, List, Protocol, Union, runtime_checkable


@runtime_checkable
class SimilaritySearchable(Protocol):
    def get_full_value(self) -> str: ...

    def get_search_terms(self) -> Iterable[str]: ...


def levenshtein_distance(s1, s2, cost_insert=1, cost_delete=1, cost_substitute=1):
    # Create a table to store results of subproblems
    m, n = len(s1), len(s2)
    dp = [[0] * (n + 1) for _ in range(m + 1)]

    # Fill the table for base cases
    for i in range(m + 1):
        dp[i][0] = i * cost_delete  # Cost of deletions to make s1 empty
    for j in range(n + 1):
        dp[0][j] = j * cost_insert  # Cost of insertions to make s1 into s2

    # Compute the distance based on the recursion described
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if s1[i - 1] == s2[j - 1]:
                cost = 0
            else:
                cost = cost_substitute

            dp[i][j] = min(
                dp[i - 1][j] + cost_delete,
                dp[i][j - 1] + cost_insert,
                dp[i - 1][j - 1] + cost,
            )

    return dp[m][n]


def score_string_similarity(
    source: SimilaritySearchable | str, search_term: str, weights=(1, 1, 1), score_cutoff=100
) -> int:
    """Calculates and returns the Levenshtein distance between a source string and a search term.

    Weights are supplied as tuple of scores to apply for the operations respectively:
        (addition_cost, deletion_cost, change_cost)

    By default every type of change costs 1.

    Experimentation has found that values of (1, 10, 10) with a cutoff of 35 makes a pretty pleasant
    set of values to use for general purpose search where the user is typing what they are looking for
    in a shorthand way (where changes / deletions are much less likely than additions).

    This makes it possible to just type a few letters of the word / phrase you are searching for, in the correct
    order, and get a low score even when a decent number of letters needed to be added. Any deletions or changes
    however will dramatically worsen the score.

    Default weights and score_cutoff found to work via manual tweaks to perform well for "partial
    string matching" where a user can search by inputing any of the letters in the source string,
    in the proper order. Deletions / substitutions of letters are heavily penalized.

    When the distance is above the cutoff, -1 is returned.
    """
    match source:
        case str():
            score = levenshtein_distance(
                search_term, source, cost_insert=weights[0], cost_delete=weights[1], cost_substitute=weights[2]
            )
            # score = distance(search_term, source, weights=weights, score_cutoff=score_cutoff)
        case SimilaritySearchable():
            # test each the terms the Searchable item provides and rank them best to worst (lowest to highest)
            term_scores = sorted(
                [
                    (
                        levenshtein_distance(
                            search_term,
                            source_term,
                            cost_insert=weights[0],
                            cost_delete=weights[1],
                            cost_substitute=weights[2],
                        ),
                        source_term,
                    )
                    for source_term in source.get_search_terms()
                ]
            )
            score = term_scores[0][0]
        case _:
            raise TypeError(f"source must be a str or implement the {SimilaritySearchable.__name__} protocol")

    return -1 if score > score_cutoff else score


@dataclass(frozen=True, order=True)
class MatchedString:
    score: int
    value: str  # the full string value
    match_term: str  # the specific term that was matched against (possibly a sub-value of "value"

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, MatchedString):
            return (self.score, self.value) == (other.score, other.value)
        else:
            raise ValueError(type(other))

    def __str__(self):
        return self.value


def get_top_n_results(
    searchable_items: List[Union[str, SimilaritySearchable]],
    search_term: str,
    max_results: int = 5,
    weights=(1, 10, 10),  # by default, heavily penalize changes/deletions
    max_distance=35,
) -> List[MatchedString]:
    """Finds and returns the items most similar to the search term from a list of searchable items."""

    matches = []
    similarity_calculator = partial(
        score_string_similarity, search_term=search_term, weights=weights, score_cutoff=max_distance
    )

    for item in searchable_items:
        if isinstance(item, str):
            similarity_score = similarity_calculator(item)
            if similarity_score > -1:
                matches.append(MatchedString(score=similarity_score, value=item, match_term=item))
        else:
            # if the item is not a string, it must implement the "SimilaritySearchable" protocol
            term_scores = sorted([(similarity_calculator(term), term) for term in item.get_search_terms()])
            try:
                winning_score, winning_term = next(x for x in term_scores if x[0] > -1)
                matches.append(MatchedString(score=winning_score, value=item.get_full_value(), match_term=winning_term))
            except StopIteration:
                continue

    return sorted(matches)[:max_results]  # noqa



from typing import Callable, List, Optional, Union

import streamlit as st
import streamlit.components.v1 as components


def item_paginator(
    title: str,
    items: Union[int, List[str]],
    item_handler_fn: Callable,
    item_actions: Optional[dict] = None,
    display_item_names=False,
    enable_keypress_nav=False,
):
    """
    This function implements a paginator for items in a Streamlit application.

    Args:
        title (str): The title of the paginator, displayed at the top.
        items (Union[int, List[str]]): If an integer is provided, it denotes the total number of items.
            If a list of strings is provided, each string is considered an item and its length denotes the
            total number of items.
        item_handler_fn (Callable): A function called to handle displaying the current item. This function should
            accept an integer as an argument, representing the current item's index.
        item_actions (Optional[dict]): An optional dictionary of actions that can be applied to the current item.
            The keys of the dictionary should be the names of the actions, and the values should be functions
            that accept the current item's index as an argument. If this argument is not provided, the paginator
            will not display any actions.
        display_item_names (bool, optional): If set to True, item names will be displayed in the selection dropdown
            instead of indices when items is a list of strings. If items is an integer, setting this to True will
            raise a ValueError. Defaults to False.
        enable_keypress_nav (bool, optional): If set to True, enables left and right key navigation to go to the
            previous or next item, respectively. Defaults to False.
    """

    item_num_var = f"ItemPaginator:{title}#item_num"
    if item_num_var not in st.session_state:
        st.session_state[item_num_var] = 0

    st.markdown(f"<h3 style='text-align: center'>{title}</h3>", unsafe_allow_html=True)
    if display_item_names:
        c1, c2, c3, c4, c6, c7 = st.columns((2, 2, 1, 3, 3, 2))
        c5 = None
    else:
        c1, c2, c3, c4, c5, c6, c7 = st.columns((2, 2, 1, 3, 1, 2, 2))
    if isinstance(items, int):
        if display_item_names:
            raise ValueError("Cannot set display_item_names=True when passing an integer for items")
    item_count = items if isinstance(items, int) else len(items)

    if st.session_state[item_num_var] > item_count - 1:
        st.session_state[item_num_var] = item_count - 1

    if st.session_state[item_num_var] < 0:
        st.session_state[item_num_var] = 0

    if not item_count:
        st.write("No items to display")
        return

    def decrement_item_index():
        st.session_state[item_num_var] = max(0, st.session_state[item_num_var] - 1)

    def set_item_index(item_num: int):
        st.session_state[item_num_var] = item_num

    c1.button(
        "Prev",
        on_click=decrement_item_index,
        key=f"ItemPaginator:{title}#Previous_btn",
        use_container_width=True,
    )

    if item_actions:
        with c2:
            selected_action = st.selectbox("select-item-action", sorted(item_actions), label_visibility="collapsed")

        if c3.button("Go", use_container_width=True, key=f"ItemPaginator:{title}#Go_btn") and selected_action:
            item_actions[selected_action](st.session_state[item_num_var])
            st.rerun()

    c4.markdown(
        f"<div style='text-align: center'>Viewing {st.session_state[item_num_var] + 1} of {item_count}</div>",
        unsafe_allow_html=True,
    )
    # hide the "Goto" label when displaying item labels in the select box, to give it more room to display
    if not display_item_names:
        c5.write("Goto:")

    with c6:

        def _format_for_select(idx):
            return items[idx] if display_item_names else idx

        selected_value = st.selectbox(
            "goto_item_number",
            range(item_count),
            format_func=_format_for_select,
            index=st.session_state[item_num_var],
            label_visibility="collapsed",
            key=f"ItemPaginator:{title}#goto_item_number_select",
        )
        if selected_value != st.session_state[item_num_var]:
            set_item_index(selected_value)
            st.rerun()

    def increment_item_index():
        st.session_state[item_num_var] = min(item_count - 1, st.session_state[item_num_var] + 1)

    c7.button(
        "Next",
        on_click=increment_item_index,
        key=f"ItemPaginator:{title}#Next_btn",
        use_container_width=True,
    )

    if enable_keypress_nav:
        enable_keypress_navigation()

    item_handler_fn(st.session_state[item_num_var])


def reset_paginator(title, st=None):
    if not st:
        import streamlit as st
    item_num_var = f"{title}#item_num"
    st.session_state[item_num_var] = 0


def enable_keypress_navigation(prev_button="Prev", next_button="Next"):
    """Enables keypress navigation in Streamlit. Allows navigation using left and right arrow keys.

    You must have existing buttons to navigate back / forth, with text matching the supplied parameter values.
    Derived from: https://github.com/streamlit/streamlit/issues/1291#issuecomment-1022408379
    """

    components.html(
        f"""
    <script>
    const doc = window.parent.document;
    buttons = Array.from(doc.querySelectorAll('button[kind=secondary]'));
    window.parent.console.log(buttons);
    const left_button = buttons.find(el => el.innerText === "{prev_button}");
    window.parent.console.log(left_button);
    const right_button = buttons.find(el => el.innerText === "{next_button}");
    window.parent.console.log(right_button);
    doc.addEventListener('keydown', function(e) {{
        switch (e.keyCode) {{
            case 37: // (37 = left arrow)
                left_button.click();
                break;
            case 39: // (39 = right arrow)
                right_button.click();
                break;
        }}
    }});
    </script>
    """,
        height=0,
        width=0,
    )



from duckduckgo_search import DDGS
from pydantic import BaseModel

from supersullytools.llm.agent import AgentTool, PydanticModel


class NewsSearch(BaseModel):
    """Search recent US news."""

    q: str


class WebSearch(BaseModel):
    q: str


class ImageSearch(BaseModel):
    q: str


def get_ddg_tools(include_news=True, include_web=True, include_image=True) -> list[AgentTool]:
    def _handle_tool_usage(params: PydanticModel):
        with DDGS() as ddgs:
            match params:
                case NewsSearch():
                    result = ddgs.news(params.q, region="us-en", safesearch="off", max_results=10)
                case WebSearch():
                    result = ddgs.text(params.q, region="us-en", safesearch="off", backend="api", max_results=10)
                case ImageSearch():
                    result = ddgs.images(
                        params.q,
                        region="us-en",
                        safesearch="off",
                        size=None,
                        # color="Monochrome",
                        type_image=None,
                        layout=None,
                        license_image=None,
                        max_results=10,
                    )
        return result

    tools = []

    def _add_tool(model, is_safe):
        tools.append(
            AgentTool(name=model.__name__, params_model=model, mechanism=_handle_tool_usage, safe_tool=is_safe)
        )

    if include_web:
        _add_tool(WebSearch, True)
    if include_news:
        _add_tool(NewsSearch, True)
    if include_image:
        _add_tool(ImageSearch, True)

    return tools


# def get_result(search_type, query) -> list[dict | str]:
#     results = ["answer"]
#
#         match search_type:
#             case "text":
#                 print(f"Searching for text {query=}")
#
#             case "answers":
#                 print(f"Searching for answers {query=}")
#                 results = ddgs.answers(query)
#             case "news":
#                 print(f"Searching for news {query=}")
#                 results = ddgs.news(
#                     query,
#                     region="us-en",
#                     safesearch="off",
#                     max_results=10,
#                 )
#             case "images":
#                 print(f"Searching for images {query=}")
#                 results = ddgs.images(
#                     query,
#                     region="wt-wt",
#                     safesearch="off",
#                     size=None,
#                     # color="Monochrome",
#                     type_image=None,
#                     layout=None,
#                     license_image=None,
#                     max_results=10,
#                 )
#             case "videos":
#                 print(f"Searching for videos {query=}")
#                 results = ddgs.videos(
#                     query,
#                     region="wt-wt",
#                     safesearch="off",
#                     max_results=10,
#                 )
#             case "maps":
#                 pass
#             case "translate":
#                 pass
#             case "suggestions":
#                 results = ddgs.suggestions(query)
#             case _:
#                 raise ValueError("Unhandled search type")
#         return list(results)
